"""Logger that writes to stderr."""

import sys

__original_stdout = None


def redirect_stdout() -> None:
    """Redirect all messages written to stdout to stderr. When you do not want
    to redirect stdout anymore, call `reset_stdout`."""

    global __original_stdout

    __original_stdout = sys.stdout
    sys.stdout = sys.stderr


def reset_stdout() -> None:
    """Reset stdout to its original value. This function should always be
    called after `redirect_stdout` to avoid unexpected behavior."""

    global __original_stdout

    if __original_stdout is None:
        sys.stdout = sys.__stdout__
        return

    sys.stdout = __original_stdout
    __original_stdout = None


def log(message: str) -> None:
    """
    Log a message to stderr.

    Args:
        message: The message to log.
    """

    print(message, file=sys.stderr)



"""Module for handling output destinations and data."""

import copy
import csv
import datetime
import json
import os
import sys
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Union

from pydantic import Field

from nextmv.base_model import BaseModel
from nextmv.logger import reset_stdout
from nextmv.options import Options


class RunStatistics(BaseModel):
    """
    Statistics about a general run.

    Parameters
    ----------
    duration : float, optional
        Duration of the run in seconds.
    iterations : int, optional
        Number of iterations.
    custom : Union[Any, dict[str, Any]], optional
        Custom statistics created by the user. Can normally expect a `dict[str,
        Any]`.
    """

    duration: Optional[float] = None
    """Duration of the run in seconds."""
    iterations: Optional[int] = None
    """Number of iterations."""
    custom: Optional[
        Union[
            Any,
            dict[str, Any],
        ]
    ] = None
    """Custom statistics created by the user. Can normally expect a `dict[str,
    Any]`."""


class ResultStatistics(BaseModel):
    """
    Statistics about a specific result.

    Parameters
    ----------
    duration : float, optional
        Duration of the run in seconds.
    value : float, optional
        Value of the result.
    custom : Union[Any, dict[str, Any]], optional
        Custom statistics created by the user. Can normally expect a `dict[str,
        Any]`.
    """

    duration: Optional[float] = None
    """Duration of the run in seconds."""
    value: Optional[float] = None
    """Value of the result."""
    custom: Optional[
        Union[
            Any,
            dict[str, Any],
        ]
    ] = None
    """Custom statistics created by the user. Can normally expect a `dict[str,
    Any]`."""


class DataPoint(BaseModel):
    """
    A data point.

    Parameters
    ----------
    x : float
        X coordinate of the data point.
    y : float
        Y coordinate of the data point.
    """

    x: float
    """X coordinate of the data point."""
    y: float
    """Y coordinate of the data point."""


class Series(BaseModel):
    """
    A series of data points.

    Parameters
    ----------
    name : str, optional
        Name of the series.
    data_points : list[DataPoint], optional
        Data of the series.
    """

    name: Optional[str] = None
    """Name of the series."""
    data_points: Optional[list[DataPoint]] = None
    """Data of the series."""


class SeriesData(BaseModel):
    """
    Data of a series.

    Parameters
    ----------
    value : Series, optional
        A series for the value of the solution.
    custom : list[Series], optional
        A list of series for custom statistics.
    """

    value: Optional[Series] = None
    """A series for the value of the solution."""
    custom: Optional[list[Series]] = None
    """A list of series for custom statistics."""


class Statistics(BaseModel):
    """
    Statistics of a solution.

    Parameters
    ----------
    run : RunStatistics, optional
        Statistics about the run.
    result : ResultStatistics, optional
        Statistics about the last result.
    series_data : SeriesData, optional
        Series data about some metric.
    statistics_schema : str, optional
        Schema (version). This class only supports `v1`.
    """

    run: Optional[RunStatistics] = None
    """Statistics about the run."""
    result: Optional[ResultStatistics] = None
    """Statistics about the last result."""
    series_data: Optional[SeriesData] = None
    """Data of the series."""
    statistics_schema: Optional[str] = Field(serialization_alias="schema", default="v1")
    """Schema (version). This class only supports `v1`."""


class OutputFormat(str, Enum):
    """Format of an `Input`."""

    JSON = "JSON"
    """JSON format, utf-8 encoded."""
    CSV_ARCHIVE = "CSV_ARCHIVE"
    """CSV archive format: multiple CSV files."""


class VisualSchema(str, Enum):
    """Schema of a visual asset."""

    CHARTJS = "chartjs"
    """Tells Nextmv Console to render the custom asset data with the Chart.js
    library."""
    GEOJSON = "geojson"
    """Tells Nextmv Console to render the custom asset data as GeoJSON on a
    map."""
    PLOTLY = "plotly"
    """Tells Nextmv Console to render the custom asset data with the Plotly
    library."""


class Visual(BaseModel):
    """
    Visual schema of an asset that defines how it is plotted in the Nextmv
    Console.
    """

    visual_schema: VisualSchema = Field(serialization_alias="schema")
    """Schema of the visual asset."""
    label: str
    """Label for the custom tab of the visual asset in the Nextmv Console."""

    visual_type: Optional[str] = Field(serialization_alias="type", default="custom-tab")
    """Defines the type of custom visual, currently there is only one type:
    `custom-tab`. This renders the visual in its own tab view of the run
    details."""

    def __post_init__(self):
        if self.visual_schema not in VisualSchema:
            raise ValueError(f"unsupported schema: {self.visual_schema}, supported schemas are {VisualSchema}")

        if self.visual_type != "custom-tab":
            raise ValueError(f"unsupported visual_type: {self.visual_type}, supported types are `custom-tab`")


class Asset(BaseModel):
    """
    An asset represents downloadable information that is part of the `Output`.
    """

    name: str
    """Name of the asset."""
    content: Any
    """Content of the asset. The type must be serializable to JSON."""

    content_type: Optional[str] = "json"
    """Content type of the asset. Only `json` is allowed"""
    description: Optional[str] = None
    """Description of the asset."""
    visual: Optional[Visual] = None
    """Visual schema of the asset."""

    def __post_init__(self):
        if self.content_type != "json":
            raise ValueError(f"unsupported content_type: {self.content_type}, supported types are `json`")


@dataclass
class Output:
    """
    Output of a decision problem. This class is used to be later be written to
    some location.

    The output can be in different formats, such as JSON (default) or
    CSV_ARCHIVE.

    If you used options, you can also include them in the output, to be
    serialized to the write location.

    The most important part of the output is the solution, which represents the
    result of the decision problem. The solution’s type must match the
    `output_format`:

    - `OutputFormat.JSON`: the data must be `dict[str, Any]`.
    - `OutputFormat.CSV_ARCHIVE`: the data must be `dict[str, list[dict[str,
      Any]]]`. The keys represent the file names where the data should be
      written. The values are lists of dictionaries, where each dictionary
      represents a row in the CSV file.

    The statistics are used to keep track of different metrics that were
    obtained after the run was completed. Although it can be a simple
    dictionary, we recommend using the `Statistics` class to ensure that the
    data is correctly formatted.

    Parameters
    ----------
    options : Options, optional
        Options that the `Input` were created with.
    output_format : OutputFormat, optional
        Format of the output data. Default is `OutputFormat.JSON`.
    solution : Union[dict[str, Any], dict[str, list[dict[str, Any]]], optional
        The solution to the decision problem.
    statistics : Union[Statistics, dict[str, Any], optional
        Statistics of the solution.
    """

    options: Optional[Options] = None
    """Options that the `Input` were created with."""
    output_format: Optional[OutputFormat] = OutputFormat.JSON
    """Format of the output data. Default is `OutputFormat.JSON`."""
    solution: Optional[
        Union[
            Union[dict[str, Any], Any],  # JSON
            dict[str, list[dict[str, Any]]],  # CSV_ARCHIVE
        ]
    ] = None
    """The solution to the decision problem."""
    statistics: Optional[Union[Statistics, dict[str, Any]]] = None
    """Statistics of the solution."""
    csv_configurations: Optional[dict[str, Any]] = None
    """Optional configuration for writing CSV files, to be used when the
    `output_format` is OutputFormat.CSV_ARCHIVE. These configurations are
    passed as kwargs to the `DictWriter` class from the `csv` module."""
    assets: Optional[list[Asset]] = None
    """Optional list of assets to be included in the output."""

    def __post_init__(self):
        """Check that the solution matches the format given to initialize the
        class."""

        # Capture a snapshot of the options that were used to create the class
        # so even if they are changed later, we have a record of the original.
        init_options = self.options
        new_options = copy.deepcopy(init_options)
        self.options = new_options

        if self.solution is None:
            return

        if self.output_format == OutputFormat.JSON:
            try:
                _ = json.dumps(self.solution, default=_custom_serial)
            except (TypeError, OverflowError) as e:
                raise ValueError(
                    f"Output has output_format OutputFormat.JSON and "
                    f"Output.solution is of type {type(self.solution)}, which is not JSON serializable"
                ) from e

        elif self.output_format == OutputFormat.CSV_ARCHIVE and not isinstance(self.solution, dict):
            raise ValueError(
                f"unsupported Output.solution type: {type(self.solution)} with "
                "output_format OutputFormat.CSV_ARCHIVE, supported type is `dict`"
            )


class OutputWriter:
    """Base class for writing outputs."""

    def write(self, output: Output, *args, **kwargs) -> None:
        """
        Write the output data. This method should be implemented by subclasses.
        """

        raise NotImplementedError


class LocalOutputWriter(OutputWriter):
    """
    Class for write outputs to local files or stdout. Call the `write` method
    to write the output data.
    """

    def _write_json(
        output: Union[Output, dict[str, Any]],
        options: dict[str, Any],
        statistics: dict[str, Any],
        assets: list[dict[str, Any]],
        path: Optional[str] = None,
    ) -> None:
        if isinstance(output, dict):
            final_output = output
        else:
            solution = output.solution if output.solution is not None else {}
            final_output = {
                "options": options,
                "solution": solution,
                "statistics": statistics,
                "assets": assets,
            }

        serialized = json.dumps(
            final_output,
            indent=2,
            default=_custom_serial,
        )

        if path is None or path == "":
            print(serialized, file=sys.stdout)
            return

        with open(path, "w", encoding="utf-8") as file:
            file.write(serialized + "\n")

    def _write_archive(
        output: Output,
        options: dict[str, Any],
        statistics: dict[str, Any],
        assets: list[dict[str, Any]],
        path: Optional[str] = None,
    ) -> None:
        dir_path = "output"
        if path is not None and path != "":
            if os.path.isfile(path):
                raise ValueError(f"The path refers to an existing file: {path}")

            dir_path = path

        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        serialized = json.dumps(
            {
                "options": options,
                "statistics": statistics,
                "assets": assets,
            },
            indent=2,
        )
        print(serialized, file=sys.stdout)

        if output.solution is None:
            return

        csv_configurations = output.csv_configurations
        if csv_configurations is None:
            csv_configurations = {}

        for file_name, data in output.solution.items():
            file_path = os.path.join(dir_path, f"{file_name}.csv")
            with open(file_path, "w", encoding="utf-8", newline="") as file:
                writer = csv.DictWriter(
                    file,
                    fieldnames=data[0].keys(),
                    **csv_configurations,
                )
                writer.writeheader()
                writer.writerows(data)

    # Callback functions for writing the output data.
    FILE_WRITERS = {
        OutputFormat.JSON: _write_json,
        OutputFormat.CSV_ARCHIVE: _write_archive,
    }

    def write(
        self,
        output: Union[Output, dict[str, Any]],
        path: Optional[str] = None,
        skip_stdout_reset: bool = False,
    ) -> None:
        """
        Write the `output` to the local filesystem. Consider the following for
        the `path` parameter, depending on the `Output.output_format`:

        - `OutputFormat.JSON`: the `path` is the file where the JSON data will
            be written. If empty or `None`, the data will be written to stdout.
        - `OutputFormat.CSV_ARCHIVE`: the `path` is the directory where the CSV
            files will be written. If empty or `None`, the data will be written
            to a directory named `output` under the current working directory.
            The `Output.options` and `Output.statistics` will be written to
            stdout.

        This function detects if stdout was redirected and resets it to avoid
        unexpected behavior. If you want to skip this behavior, set the
        `skip_stdout_reset` parameter to `True`.

        If the `output` is a `dict`, it will be simply written to the specified
        `path`, as a passthrough. On the other hand, if the `output` is of type
        `Output`, a more structured object will be written, which adheres to
        the schema specified by the corresponding `Output` class.

        Parameters
        ----------
        output: Output, dict[str, Any]
            Output data to write.
        path : str
            Path to write the output data to.
        skip_stdout_reset : bool, optional
            Skip resetting stdout before writing the output data. Default is
            `False`.

        Raises
        ------
        ValueError
            If the `Output.output_format` is not supported.
        """

        # If the user forgot to reset stdout after redirecting it, we need to
        # do it here to avoid unexpected behavior.
        if sys.stdout is not sys.__stdout__ and not skip_stdout_reset:
            reset_stdout()

        if isinstance(output, Output):
            output_format = output.output_format
        elif isinstance(output, dict):
            output_format = OutputFormat.JSON
        else:
            raise TypeError(f"unsupported output type: {type(output)}, supported types are `Output` or `dict`")

        statistics = self._extract_statistics(output)
        options = self._extract_options(output)
        assets = self._extract_assets(output)

        self.FILE_WRITERS[output_format](
            output=output,
            options=options,
            statistics=statistics,
            assets=assets,
            path=path,
        )

    @staticmethod
    def _extract_statistics(output: Union[Output, dict[str, Any]]) -> dict[str, Any]:
        """Extract JSON-serializable statistics."""

        statistics = {}

        if not isinstance(output, Output):
            return statistics

        stats = output.statistics

        if stats is None:
            return statistics

        if isinstance(stats, Statistics):
            statistics = stats.to_dict()
        elif isinstance(stats, dict):
            statistics = stats
        else:
            raise TypeError(f"unsupported statistics type: {type(stats)}, supported types are `Statistics` or `dict`")

        return statistics

    @staticmethod
    def _extract_options(output: Union[Output, dict[str, Any]]) -> dict[str, Any]:
        """Extract JSON-serializable options."""

        options = {}

        if not isinstance(output, Output):
            return options

        opt = output.options

        if opt is None:
            return options

        if isinstance(opt, Options):
            options = opt.to_dict()
        elif isinstance(opt, dict):
            options = opt
        else:
            raise TypeError(f"unsupported options type: {type(opt)}, supported types are `Options` or `dict`")

        return options

    @staticmethod
    def _extract_assets(output: Union[Output, dict[str, Any]]) -> list[dict[str, Any]]:
        """Extract JSON-serializable assets."""

        assets = []

        if not isinstance(output, Output):
            return assets

        assts = output.assets

        if assts is None:
            return assets

        for ix, asset in enumerate(assts):
            if isinstance(asset, Asset):
                assets.append(asset.to_dict())
            elif isinstance(asset, dict):
                assets.append(asset)
            else:
                raise TypeError(f"unsupported asset {ix}, type: {type(asset)}; supported types are `Asset` or `dict`")

        return assets


def write_local(
    output: Union[Output, dict[str, Any]],
    path: Optional[str] = None,
    skip_stdout_reset: bool = False,
) -> None:
    """
    This is a convenience function for instantiating a `LocalOutputWriter` and
    calling its `write` method.

    Write the `output` to the local filesystem. Consider the following for the
    `path` parameter, depending on the `Output.output_format`:

    - `OutputFormat.JSON`: the `path` is the file where the JSON data will
        be written. If empty or `None`, the data will be written to stdout.
    - `OutputFormat.CSV_ARCHIVE`: the `path` is the directory where the CSV
        files will be written. If empty or `None`, the data will be written
        to a directory named `output` under the current working directory.
        The `Output.options` and `Output.statistics` will be written to
        stdout.

    This function detects if stdout was redirected and resets it to avoid
    unexpected behavior. If you want to skip this behavior, set the
    `skip_stdout_reset` parameter to `True`.

    Parameters
    ----------
    output : Output, dict[str, Any]
        Output data to write.
    path : str
        Path to write the output data to.
    skip_stdout_reset : bool, optional
        Skip resetting stdout before writing the output data. Default is
        `False`.

    Raises
    ------
    ValueError
        If the `Output.output_format` is not supported.
    """

    writer = LocalOutputWriter()
    writer.write(output, path, skip_stdout_reset)


def _custom_serial(obj: Any):
    """JSON serializer for objects not serializable by default one."""

    if isinstance(obj, (datetime.datetime | datetime.date)):
        return obj.isoformat()

    raise TypeError(f"Type {type(obj)} not serializable")



"""This module contains the application class."""

import json
import shutil
import time
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional, Union

import requests

from nextmv.base_model import BaseModel
from nextmv.cloud import package
from nextmv.cloud.acceptance_test import AcceptanceTest, ExperimentStatus, Metric
from nextmv.cloud.batch_experiment import BatchExperiment, BatchExperimentMetadata, BatchExperimentRun
from nextmv.cloud.client import Client, get_size
from nextmv.cloud.input_set import InputSet
from nextmv.cloud.instance import Configuration, Instance
from nextmv.cloud.manifest import Manifest
from nextmv.cloud.status import Status, StatusV2
from nextmv.cloud.version import Version
from nextmv.logger import log
from nextmv.model import Model, ModelConfiguration

_MAX_RUN_SIZE: int = 5 * 1024 * 1024
"""Maximum size of the run input/output. This value is used to determine
whether to use the large input upload and/or result download endpoints."""


class DownloadURL(BaseModel):
    """Result of getting a download URL."""

    url: str
    """URL to use for downloading the file."""


class ErrorLog(BaseModel):
    """Error log of a run, when it was not successful."""

    error: Optional[str] = None
    """Error message."""
    stdout: Optional[str] = None
    """Standard output."""
    stderr: Optional[str] = None
    """Standard error."""


class Metadata(BaseModel):
    """Metadata of a run, whether it was successful or not."""

    application_id: str
    """ID of the application where the run was submitted to."""
    application_instance_id: str
    """ID of the instance where the run was submitted to."""
    application_version_id: str
    """ID of the version of the application where the run was submitted to."""
    created_at: datetime
    """Date and time when the run was created."""
    duration: float
    """Duration of the run in milliseconds."""
    error: str
    """Error message if the run failed."""
    input_size: float
    """Size of the input in bytes."""
    output_size: float
    """Size of the output in bytes."""
    status: Status
    """Deprecated: use status_v2."""
    status_v2: StatusV2
    """Status of the run."""


class PollingOptions(BaseModel):
    """Options to use when polling for a run result."""

    backoff: float = 1
    """Backoff factor to use between polls. Leave this at 1 to poll at a
    constant rate."""
    delay: float = 1
    """Delay to use between polls, in seconds."""
    initial_delay: float = 1
    """Initial delay to use before starting the polling strategy, in
    seconds."""
    max_delay: float = 20
    """Maximum delay to use between polls, in seconds. This parameter is
    activated when the backoff parameter is greater than 1, such that the delay
    is increasing after each poll."""
    max_duration: float = 300
    """Maximum duration of the polling strategy, in seconds."""
    max_tries: int = 20
    """Maximum number of tries to use."""


_DEFAULT_POLLING_OPTIONS: PollingOptions = PollingOptions()
"""Default polling options to use when polling for a run result."""


class RunInformation(BaseModel):
    """Information of a run."""

    description: str
    """Description of the run."""
    id: str
    """ID of the run."""
    metadata: Metadata
    """Metadata of the run."""
    name: str
    """Name of the run."""
    user_email: str
    """Email of the user who submitted the run."""


class RunResult(RunInformation):
    """Result of a run, whether it was successful or not."""

    error_log: Optional[ErrorLog] = None
    """Error log of the run. Only available if the run failed."""
    output: Optional[dict[str, Any]] = None
    """Output of the run. Only available if the run succeeded."""


class RunLog(BaseModel):
    """Log of a run."""

    log: str
    """Log of the run."""


class UploadURL(BaseModel):
    """Result of getting an upload URL."""

    upload_id: str
    """ID of the upload."""
    upload_url: str
    """URL to use for uploading the file."""


@dataclass
class Application:
    """An application is a published decision model that can be executed."""

    client: Client
    """Client to use for interacting with the Nextmv Cloud API."""
    id: str
    """ID of the application."""

    default_instance_id: str = "devint"
    """Default instance ID to use for submitting runs."""
    endpoint: str = "v1/applications/{id}"
    """Base endpoint for the application."""
    experiments_endpoint: str = "{base}/experiments"
    """Base endpoint for the experiments in the application."""

    def __post_init__(self):
        """Logic to run after the class is initialized."""

        self.endpoint = self.endpoint.format(id=self.id)
        self.experiments_endpoint = self.experiments_endpoint.format(base=self.endpoint)

    def acceptance_test(self, acceptance_test_id: str) -> AcceptanceTest:
        """
        Get an acceptance test.

        Args:
            acceptance_test_id: ID of the acceptance test.

        Returns:
            Acceptance test.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.experiments_endpoint}/acceptance/{acceptance_test_id}",
        )

        return AcceptanceTest.from_dict(response.json())

    def batch_experiment(self, batch_id: str) -> BatchExperiment:
        """
        Get a batch experiment.

        Args:
            batch_id: ID of the batch experiment.

        Returns:
            Batch experiment.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.experiments_endpoint}/batch/{batch_id}",
        )

        return BatchExperiment.from_dict(response.json())

    def cancel_run(self, run_id: str) -> None:
        """
        Cancel a run.

        Args:
            run_id: ID of the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        _ = self.client.request(
            method="PATCH",
            endpoint=f"{self.endpoint}/runs/{run_id}/cancel",
        )

    def delete_batch_experiment(self, batch_id: str) -> None:
        """
        Deletes a batch experiment, along with all the associated information,
        such as its runs.

        Args:
            batch_id: ID of the batch experiment.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        _ = self.client.request(
            method="DELETE",
            endpoint=f"{self.experiments_endpoint}/batch/{batch_id}",
        )

    def delete_acceptance_test(self, acceptance_test_id: str) -> None:
        """
        Deletes an acceptance test, along with all the associated information
        such as the underlying batch experiment.

        Args:
            acceptance_test_id: ID of the acceptance test.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        _ = self.client.request(
            method="DELETE",
            endpoint=f"{self.experiments_endpoint}/acceptance/{acceptance_test_id}",
        )

    def input_set(self, input_set_id: str) -> InputSet:
        """
        Get an input set.

        Args:
            input_set_id: ID of the input set.

        Returns:
            Input set.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.experiments_endpoint}/inputsets/{input_set_id}",
        )

        return InputSet.from_dict(response.json())

    def instance(self, instance_id: str) -> Instance:
        """
        Get an instance.

        Args:
            instance_id: ID of the instance.

        Returns:
            Instance.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/instances/{instance_id}",
        )

        return Instance.from_dict(response.json())

    def list_acceptance_tests(self) -> list[AcceptanceTest]:
        """
        List all acceptance tests.

        Returns:
            List of acceptance tests.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.experiments_endpoint}/acceptance",
        )

        return [AcceptanceTest.from_dict(acceptance_test) for acceptance_test in response.json()]

    def list_batch_experiments(self) -> list[BatchExperimentMetadata]:
        """
        List all batch experiments.

        Returns:
            List of batch experiments.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.experiments_endpoint}/batch",
        )

        return [BatchExperimentMetadata.from_dict(batch_experiment) for batch_experiment in response.json()]

    def list_input_sets(self) -> list[InputSet]:
        """
        List all input sets.

        Returns:
            List of input sets.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.experiments_endpoint}/inputsets",
        )

        return [InputSet.from_dict(input_set) for input_set in response.json()]

    def list_instances(self) -> list[Instance]:
        """
        List all instances.

        Returns:
            List of instances.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/instances",
        )

        return [Instance.from_dict(instance) for instance in response.json()]

    def list_versions(self) -> list[Version]:
        """
        List all versions.

        Returns:
            List of versions.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/versions",
        )

        return [Version.from_dict(version) for version in response.json()]

    def new_acceptance_test(
        self,
        candidate_instance_id: str,
        baseline_instance_id: str,
        id: str,
        metrics: list[Union[Metric, dict[str, Any]]],
        name: str,
        input_set_id: Optional[str] = None,
        description: Optional[str] = None,
    ) -> AcceptanceTest:
        """
        Create a new acceptance test. The acceptance test is based on a batch
        experiment. If you already started a batch experiment, you don't need
        to provide the input_set_id parameter. In that case, the ID of the
        acceptance test and the batch experiment must be the same. If the batch
        experiment does not exist, you can provide the input_set_id parameter
        and a new batch experiment will be created for you.

        Args:
            candidate_instance_id: ID of the candidate instance.
            baseline_instance_id: ID of the baseline instance.
            id: ID of the acceptance test.
            metrics: List of metrics to use for the acceptance test.
            name: Name of the acceptance test.
            input_set_id: ID of the input set to use for the underlying batch
                experiment, in case it hasn't been started.
            description: Description of the acceptance test.

        Returns:
            Acceptance test.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
            ValueError: If the batch experiment ID does not match the
                acceptance test ID.
        """

        if input_set_id is None:
            try:
                batch_experiment = self.batch_experiment(batch_id=id)
                batch_experiment_id = batch_experiment.id
            except requests.HTTPError as e:
                if e.response.status_code != 404:
                    raise e

                raise ValueError(
                    f"batch experiment {id} does not exist, input_set_id must be defined to create a new one"
                ) from e
        else:
            batch_experiment_id = self.new_batch_experiment(
                name=name,
                input_set_id=input_set_id,
                instance_ids=[candidate_instance_id, baseline_instance_id],
                description=description,
                id=id,
            )

        if batch_experiment_id != id:
            raise ValueError(f"batch experiment_id ({batch_experiment_id}) does not match acceptance test id ({id})")

        payload_metrics = [{}] * len(metrics)
        for i, metric in enumerate(metrics):
            payload_metrics[i] = metric.to_dict() if isinstance(metric, Metric) else metric

        payload = {
            "candidate": {"instance_id": candidate_instance_id},
            "control": {"instance_id": baseline_instance_id},
            "metrics": payload_metrics,
            "experiment_id": batch_experiment_id,
            "name": name,
        }
        if description is not None:
            payload["description"] = description
        if id is not None:
            payload["id"] = id

        response = self.client.request(
            method="POST",
            endpoint=f"{self.experiments_endpoint}/acceptance",
            payload=payload,
        )

        return AcceptanceTest.from_dict(response.json())

    def new_acceptance_test_with_result(
        self,
        candidate_instance_id: str,
        baseline_instance_id: str,
        id: str,
        metrics: list[Union[Metric, dict[str, Any]]],
        name: str,
        input_set_id: Optional[str] = None,
        description: Optional[str] = None,
        polling_options: PollingOptions = _DEFAULT_POLLING_OPTIONS,
    ) -> AcceptanceTest:
        """
        Create a new acceptance test and poll for the result. This is a
        convenience method that combines the new_acceptance_test with polling
        logic to check when the acceptance test is done.

        Args:
            candidate_instance_id: ID of the candidate instance.
            baseline_instance_id: ID of the baseline instance.
            id: ID of the acceptance test.
            metrics: List of metrics to use for the acceptance test.
            name: Name of the acceptance test.
            input_set_id: ID of the input set to use for the underlying batch
                experiment, in case it hasn't been started.
            description: Description of the acceptance test.
            polling_options: Options to use when polling for the run result.

        Returns:
            Result of the acceptance test.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
            TimeoutError: If the acceptance test does not succeed after the
                polling strategy is exhausted based on time duration.
            RuntimeError: If the acceptance test does not succeed after the
                polling strategy is exhausted based on number of tries.
        """
        _ = self.new_acceptance_test(
            candidate_instance_id=candidate_instance_id,
            baseline_instance_id=baseline_instance_id,
            id=id,
            metrics=metrics,
            name=name,
            input_set_id=input_set_id,
            description=description,
        )

        time.sleep(polling_options.initial_delay)
        delay = polling_options.delay
        polling_ok = False
        for _ in range(polling_options.max_tries):
            test_information = self.acceptance_test(acceptance_test_id=id)
            if test_information.status in [
                ExperimentStatus.completed,
                ExperimentStatus.failed,
                ExperimentStatus.canceled,
            ]:
                polling_ok = True
                break

            if delay > polling_options.max_duration:
                raise TimeoutError(
                    f"acceptance_test {id} did not succeed after {delay} seconds",
                )

            sleep_duration = min(delay, polling_options.max_delay)
            time.sleep(sleep_duration)
            delay *= polling_options.backoff

        if not polling_ok:
            raise RuntimeError(
                f"acceptance_test {id} did not succeed after {polling_options.max_tries} tries",
            )

        return test_information

    def new_batch_experiment(
        self,
        name: str,
        input_set_id: str,
        instance_ids: list[str] = None,
        description: Optional[str] = None,
        id: Optional[str] = None,
        option_sets: Optional[dict[str, dict[str, str]]] = None,
        runs: Optional[list[Union[BatchExperimentRun, dict[str, Any]]]] = None,
    ) -> str:
        """
        Create a new batch experiment.

        Args:
            name: Name of the batch experiment.
            input_set_id: ID of the input set to use for the experiment.
            instance_ids: List of instance IDs to use for the experiment.
            description: Description of the batch experiment.
            id: ID of the batch experiment.
            option_sets: Option sets to use for the experiment.
            runs: Runs to use for the experiment.

        Returns:
            ID of the batch experiment.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        payload = {
            "name": name,
            "input_set_id": input_set_id,
            "instance_ids": instance_ids,
        }
        if description is not None:
            payload["description"] = description
        if id is not None:
            payload["id"] = id
        if option_sets is not None:
            payload["option_sets"] = option_sets
        if runs is not None:
            payload_runs = [{}] * len(runs)
            for i, run in enumerate(runs):
                payload_runs[i] = run.to_dict() if isinstance(run, BatchExperimentRun) else run
            payload["runs"] = payload_runs

        response = self.client.request(
            method="POST",
            endpoint=f"{self.experiments_endpoint}/batch",
            payload=payload,
        )

        return response.json()["id"]

    def new_input_set(
        self,
        id: str,
        name: str,
        description: Optional[str] = None,
        end_time: Optional[datetime] = None,
        instance_id: Optional[str] = None,
        maximum_runs: Optional[int] = None,
        run_ids: Optional[list[str]] = None,
        start_time: Optional[datetime] = None,
    ) -> InputSet:
        """
        Create a new input set.

        Args:
            id: ID of the input set.
            name: Name of the input set.
            description: Description of the input set.
            end_time: End time of the runs to construct the input set.
            instance_id: ID of the instance to use for the input set. If not
                provided, the default_instance_id will be used.
            maximum_runs: Maximum number of runs to use for the input set.
            run_ids: IDs of the runs to use for the input set.
            start_time: Start time of the runs to construct the input set.

        Returns:
            Input set.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        payload = {
            "id": id,
            "name": name,
        }
        if description is not None:
            payload["description"] = description
        if end_time is not None:
            payload["end_time"] = end_time.isoformat()
        if instance_id is not None:
            payload["instance_id"] = instance_id
        if maximum_runs is not None:
            payload["maximum_runs"] = maximum_runs
        if run_ids is not None:
            payload["run_ids"] = run_ids
        if start_time is not None:
            payload["start_time"] = start_time.isoformat()

        response = self.client.request(
            method="POST",
            endpoint=f"{self.experiments_endpoint}/inputsets",
            payload=payload,
        )

        return InputSet.from_dict(response.json())

    def new_run(  # noqa: C901 # Lot of if statements, but clear logic.
        self,
        input: Union[dict[str, Any], BaseModel, str] = None,
        instance_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        upload_id: Optional[str] = None,
        options: Optional[dict[str, str]] = None,
        configuration: Optional[Configuration] = None,
    ) -> str:
        """
        Submit an input to start a new run of the application. Returns the
        run_id of the submitted run.

        Args:
            input: Input to use for the run. This can be JSON (given as dict
            or BaseModel) or text (given as str).
            instance_id: ID of the instance to use for the run. If not
                provided, the default_instance_id will be used.
            name: Name of the run.
            description: Description of the run.
            upload_id: ID to use when running a large input.
            options: Options to use for the run.
            configuration: Configuration to use for the run.

        Returns:
            ID of the submitted run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        input_size = 0
        if isinstance(input, BaseModel):
            input = input.to_dict()
            if input is not None:
                input_size = get_size(input)
        elif isinstance(input, dict):
            input_size = get_size(input)

        upload_url_required = isinstance(input, str) or input_size > _MAX_RUN_SIZE

        upload_id_used = upload_id is not None
        if not upload_id_used and upload_url_required:
            upload_url = self.upload_url()
            self.upload_large_input(input=input, upload_url=upload_url)
            upload_id = upload_url.upload_id
            upload_id_used = True

        if options is not None:
            for key, value in options.items():
                if not isinstance(value, str):
                    options[key] = json.dumps(value)

        payload = {}
        if upload_id_used:
            payload["upload_id"] = upload_id
        else:
            payload["input"] = input

        if name is not None:
            payload["name"] = name
        if description is not None:
            payload["description"] = description
        if options is not None:
            payload["options"] = options
        if configuration is not None:
            payload["configuration"] = configuration.to_dict()

        query_params = {
            "instance_id": instance_id if instance_id is not None else self.default_instance_id,
        }
        response = self.client.request(
            method="POST",
            endpoint=f"{self.endpoint}/runs",
            payload=payload,
            query_params=query_params,
        )

        return response.json()["run_id"]

    def new_run_with_result(
        self,
        input: Union[dict[str, Any], BaseModel] = None,
        instance_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        upload_id: Optional[str] = None,
        run_options: Optional[dict[str, str]] = None,
        polling_options: PollingOptions = _DEFAULT_POLLING_OPTIONS,
        configuration: Optional[Configuration] = None,
    ) -> RunResult:
        """
        Submit an input to start a new run of the application and poll for the
        result. This is a convenience method that combines the new_run and
        run_result_with_polling methods, applying polling logic to check when
        the run succeeded.

         Args:
            input: Input to use for the run.
            instance_id: ID of the instance to use for the run. If not
                provided, the default_instance_id will be used.
            name: Name of the run.
            description: Description of the run.
            upload_id: ID to use when running a large input.
            run_options: Options to use for the run.
            polling_options: Options to use when polling for the run result.
            configuration: Configuration to use for the run.

         Returns:
            Result of the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
            TimeoutError: If the run does not succeed after the polling
                strategy is exhausted based on time duration.
            RuntimeError: If the run does not succeed after the polling
                strategy is exhausted based on number of tries.
        """

        run_id = self.new_run(
            input=input,
            instance_id=instance_id,
            name=name,
            description=description,
            upload_id=upload_id,
            options=run_options,
            configuration=configuration,
        )

        return self.run_result_with_polling(
            run_id=run_id,
            polling_options=polling_options,
        )

    def new_version(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Version:
        """
        Create a new version using the current dev binary.

        Args:
            id: ID of the version. Will be generated if not provided.
            name: Name of the version. Will be generated if not provided.
            description: Description of the version. Will be generated if not provided.

        Returns:
            Version.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        payload = {}

        if id is not None:
            payload["id"] = id
        if name is not None:
            payload["name"] = name
        if description is not None:
            payload["description"] = description

        response = self.client.request(
            method="POST",
            endpoint=f"{self.endpoint}/versions",
            payload=payload,
        )

        return Version.from_dict(response.json())

    def new_instance(
        self,
        version_id: str,
        id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        configuration: Optional[Configuration] = None,
    ) -> Instance:
        """
        Create a new instance and associate it with a version.

        Args:
            version_id: ID of the version to associate the instance with.
            id: ID of the instance. Will be generated if not provided.
            name: Name of the instance. Will be generated if not provided.
            description: Description of the instance. Will be generated if not provided.
            configuration: Configuration to use for the instance.

        Returns:
            Instance.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        payload = {
            "version_id": version_id,
        }

        if id is not None:
            payload["id"] = id
        if name is not None:
            payload["name"] = name
        if description is not None:
            payload["description"] = description
        if configuration is not None:
            payload["configuration"] = configuration.to_dict()

        response = self.client.request(
            method="POST",
            endpoint=f"{self.endpoint}/instances",
            payload=payload,
        )

        return Instance.from_dict(response.json())

    def push(
        self,
        manifest: Optional[Manifest] = None,
        app_dir: Optional[str] = None,
        verbose: bool = False,
        model: Optional[Model] = None,
        model_configuration: Optional[ModelConfiguration] = None,
    ) -> None:
        """
        Push an app to Nextmv Cloud.

        If the manifest is not provided, an `app.yaml` file will be searched
        for in the provided path. If there is no manifest file found, an
        exception will be raised.

        There are two ways to push an app to Nextmv Cloud:
        1. Specifying `app_dir`, which is the path to an app’s root directory.
        This acts as an external strategy, where the app is composed of files
        in a directory and those apps are packaged and pushed to Nextmv Cloud.
        2. Specifying a `model` and `model_configuration`. This acts as an
        internal (or Python-native) strategy, where the app is actually a
        `nextmv.Model`. The model is encoded, some dependencies and
        accompanying files are packaged, and the app is pushed to Nextmv Cloud.

        Examples
        -------

        1. Push an app using an external strategy, i.e., specifying the app’s
        directory:
        ```python
        import os

        from nextmv import cloud

        client = cloud.Client(api_key=os.getenv("NEXTMV_API_KEY"))
        app = cloud.Application(client=client, id="<YOUR-APP-ID>")
        app.push()  # Use verbose=True for step-by-step output.
        ```

        2. Push an app using an internal strategy, i.e., specifying the model
        and model configuration:
        ```python
        import os

        import nextroute

        import nextmv
        import nextmv.cloud


        # Define the model that makes decisions. This model uses the Nextroute
        # library to solve a vehicle routing problem.
        class DecisionModel(nextmv.Model):
            def solve(self, input: nextmv.Input) -> nextmv.Output:
                nextroute_input = nextroute.schema.Input.from_dict(input.data)
                nextroute_options = nextroute.Options.extract_from_dict(input.options.to_dict())
                nextroute_output = nextroute.solve(nextroute_input, nextroute_options)

                return nextmv.Output(
                    options=input.options,
                    solution=nextroute_output.solutions[0].to_dict(),
                    statistics=nextroute_output.statistics.to_dict(),
                )


        # Define the options that the model needs.
        parameters = []
        default_options = nextroute.Options()
        for name, default_value in default_options.to_dict().items():
            parameters.append(nextmv.Parameter(name.lower(), type(default_value), default_value, name, False))

        options = nextmv.Options(*parameters)

        # Instantiate the model and model configuration.
        model = DecisionModel()
        model_configuration = nextmv.ModelConfiguration(
            name="python_nextroute_model",
            requirements=[
                "nextroute==1.8.1",
                "nextmv==0.14.0.dev1",
            ],
            options=options,
        )

        # Define the Nextmv application and push the model to the cloud.
        client = cloud.Client(api_key=os.getenv("NEXTMV_API_KEY"))
        app = cloud.Application(client=client, id="<YOUR-APP-ID>")
        manifest = nextmv.cloud.default_python_manifest()
        app.push(
            manifest=manifest,
            verbose=True,
            model=model,
            model_configuration=model_configuration,
        )
        ```

        Parameters
        ----------
        manifest : Optional[Manifest], optional
            The manifest for the app, by default None.
        app_dir : Optional[str], optional
            The path to the app’s directory, by default None.
        verbose : bool, optional
            Whether to print verbose output, by default False.
        """

        if verbose:
            log("💽 Starting build for Nextmv application.")

        if app_dir is None or app_dir == "":
            app_dir = "."

        if manifest is None:
            manifest = Manifest.from_yaml(app_dir)

        if model is not None and not isinstance(model, Model):
            raise TypeError("model must be an instance of nextmv.Model")

        if model_configuration is not None and not isinstance(model_configuration, ModelConfiguration):
            raise TypeError("model_configuration must be an instance of nextmv.ModelConfiguration")

        if (model is None and model_configuration is not None) or (model is not None and model_configuration is None):
            raise ValueError("model and model_configuration must be provided together")

        package._run_build_command(app_dir, manifest.build, verbose)
        package._run_pre_push_command(app_dir, manifest.pre_push, verbose)
        tar_file, output_dir = package._package(app_dir, manifest, model, model_configuration, verbose)
        self.__update_app_binary(tar_file, manifest, verbose)

        try:
            shutil.rmtree(output_dir)
        except OSError as e:
            raise Exception(f"error deleting output directory: {e}") from e

    def run_input(self, run_id: str) -> dict[str, Any]:
        """
        Get the input of a run.

        Args:
            run_id: ID of the run.

        Returns:
            Input of the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """
        run_information = self.run_metadata(run_id=run_id)

        query_params = None
        large = False
        if run_information.metadata.input_size > _MAX_RUN_SIZE:
            query_params = {"format": "url"}
            large = True

        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/runs/{run_id}/input",
            query_params=query_params,
        )
        if not large:
            return response.json()

        download_url = DownloadURL.from_dict(response.json())
        download_response = self.client.request(
            method="GET",
            endpoint=download_url.url,
            headers={"Content-Type": "application/json"},
        )

        return download_response.json()

    def run_logs(self, run_id: str) -> RunLog:
        """
        Get the logs of a run.

        Args:
            run_id: ID of the run.

        Returns:
            Logs of the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """
        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/runs/{run_id}/logs",
        )
        return RunLog.from_dict(response.json())

    def run_metadata(self, run_id: str) -> RunInformation:
        """
        Get the metadata of a run. The result does not include the run output.

        Args:
            run_id: ID of the run.

        Returns:
            Metadata of the run (Run result with no output).

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/runs/{run_id}/metadata",
        )

        return RunInformation.from_dict(response.json())

    def run_result(self, run_id: str) -> RunResult:
        """
        Get the result of a run. The result includes the run output.

        Args:
            run_id: ID of the run.

        Returns:
            Result of the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        run_information = self.run_metadata(run_id=run_id)

        return self.__run_result(run_id=run_id, run_information=run_information)

    def run_result_with_polling(
        self,
        run_id: str,
        polling_options: PollingOptions = _DEFAULT_POLLING_OPTIONS,
    ) -> RunResult:
        """
        Get the result of a run. The result includes the run output. This
        method polls for the result until the run finishes executing or the
        polling strategy is exhausted.

        Args:
            run_id: ID of the run.
            polling_options: Options to use when polling for the run result.

        Returns:
            Result of the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        time.sleep(polling_options.initial_delay)
        delay = polling_options.delay
        polling_ok = False
        for _ in range(polling_options.max_tries):
            run_information = self.run_metadata(run_id=run_id)
            if run_information.metadata.status_v2 in [
                StatusV2.succeeded,
                StatusV2.failed,
                StatusV2.canceled,
            ]:
                polling_ok = True
                break

            if delay > polling_options.max_duration:
                raise TimeoutError(
                    f"run {run_id} did not succeed after {delay} seconds",
                )

            sleep_duration = min(delay, polling_options.max_delay)
            time.sleep(sleep_duration)
            delay *= polling_options.backoff

        if not polling_ok:
            raise RuntimeError(
                f"run {run_id} did not succeed after {polling_options.max_tries} tries",
            )

        return self.__run_result(run_id=run_id, run_information=run_information)

    def update_instance(
        self,
        id: str,
        name: str,
        version_id: Optional[str] = None,
        description: Optional[str] = None,
        configuration: Optional[Configuration] = None,
    ) -> Instance:
        """
        Update an instance.

        Args:
            id: ID of the instance to update.
            version_id: ID of the version to associate the instance with.
            name: Name of the instance.
            description: Description of the instance.
            configuration: Configuration to use for the instance.

        Returns:
            Instance.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        payload = {}

        if version_id is not None:
            payload["version_id"] = version_id
        if name is not None:
            payload["name"] = name
        if description is not None:
            payload["description"] = description
        if configuration is not None:
            payload["configuration"] = configuration.to_dict()

        response = self.client.request(
            method="PUT",
            endpoint=f"{self.endpoint}/instances/{id}",
            payload=payload,
        )

        return Instance.from_dict(response.json())

    def upload_large_input(
        self,
        input: Union[dict[str, Any], str],
        upload_url: UploadURL,
    ) -> None:
        """
        Upload the file located at the given path to the provided upload URL.

        Args:
            upload_url: Upload URL to use for uploading the file.
            input: Input to use for the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        if isinstance(input, dict):
            input = json.dumps(input)

        _ = self.client.upload_to_presigned_url(
            url=upload_url.upload_url,
            data=input,
        )

    def upload_url(self) -> UploadURL:
        """
        Get an upload URL to use for uploading a file.

        Returns:
            Result of getting an upload URL.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="POST",
            endpoint=f"{self.endpoint}/runs/uploadurl",
        )

        return UploadURL.from_dict(response.json())

    def version(self, version_id: str) -> Version:
        """
        Get a version.

        Args:
            version_id: ID of the version.

        Returns:
            Version.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """

        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/versions/{version_id}",
        )

        return Version.from_dict(response.json())

    def __run_result(
        self,
        run_id: str,
        run_information: RunInformation,
    ) -> RunResult:
        """
        Get the result of a run. The result includes the run output. This is a
        private method that is the base for retrieving a run result, regardless
        of polling.

        Args:
            run_id: ID of the run.
            run_information: Information of the run.

        Returns:
            Result of the run.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """
        query_params = None
        large_output = False
        if run_information.metadata.output_size > _MAX_RUN_SIZE:
            query_params = {"format": "url"}
            large_output = True

        response = self.client.request(
            method="GET",
            endpoint=f"{self.endpoint}/runs/{run_id}",
            query_params=query_params,
        )
        result = RunResult.from_dict(response.json())
        if not large_output:
            return result

        download_url = DownloadURL.from_dict(response.json()["output"])
        download_response = self.client.request(
            method="GET",
            endpoint=download_url.url,
            headers={"Content-Type": "application/json"},
        )
        result.output = download_response.json()

        return result

    def __update_app_binary(
        self,
        tar_file: str,
        manifest: Manifest,
        verbose: bool = False,
    ) -> None:
        """Updates the application binary in Cloud."""

        if verbose:
            log(f'🌟 Pushing to application: "{self.id}".')

        endpoint = f"{self.endpoint}/binary"
        response = self.client.request(
            method="GET",
            endpoint=endpoint,
        )
        upload_url = response.json()["upload_url"]

        with open(tar_file, "rb") as f:
            response = self.client.request(
                method="PUT",
                endpoint=upload_url,
                data=f,
                headers={"Content-Type": "application/octet-stream"},
            )

        activation_request = {
            "requirements": {
                "executable_type": manifest.type,
                "runtime": manifest.runtime,
            },
        }
        response = self.client.request(
            method="PUT",
            endpoint=endpoint,
            payload=activation_request,
        )

        if verbose:
            log(f'💥️ Successfully pushed to application: "{self.id}".')
            log(
                json.dumps(
                    {
                        "app_id": self.id,
                        "endpoint": self.client.url,
                        "instance_url": f"{self.endpoint}/runs?instance_id=devint",
                    },
                    indent=2,
                )
            )



"""Functionality for interacting with the Nextmv Cloud."""

from .acceptance_test import AcceptanceTest as AcceptanceTest
from .acceptance_test import AcceptanceTestResults as AcceptanceTestResults
from .acceptance_test import Comparison as Comparison
from .acceptance_test import ComparisonInstance as ComparisonInstance
from .acceptance_test import DistributionPercentiles as DistributionPercentiles
from .acceptance_test import DistributionSummaryStatistics as DistributionSummaryStatistics
from .acceptance_test import ExperimentStatus as ExperimentStatus
from .acceptance_test import Metric as Metric
from .acceptance_test import MetricParams as MetricParams
from .acceptance_test import MetricResult as MetricResult
from .acceptance_test import MetricStatistics as MetricStatistics
from .acceptance_test import MetricTolerance as MetricTolerance
from .acceptance_test import MetricType as MetricType
from .acceptance_test import ResultStatistics as ResultStatistics
from .acceptance_test import StatisticType as StatisticType
from .acceptance_test import ToleranceType as ToleranceType
from .account import Account as Account
from .account import Queue as Queue
from .account import QueuedRun as QueuedRun
from .application import Application as Application
from .application import DownloadURL as DownloadURL
from .application import ErrorLog as ErrorLog
from .application import Metadata as Metadata
from .application import PollingOptions as PollingOptions
from .application import RunInformation as RunInformation
from .application import RunResult as RunResult
from .application import UploadURL as UploadURL
from .batch_experiment import BatchExperiment as BatchExperiment
from .batch_experiment import BatchExperimentInformation as BatchExperimentInformation
from .batch_experiment import BatchExperimentMetadata as BatchExperimentMetadata
from .batch_experiment import BatchExperimentRun as BatchExperimentRun
from .client import Client as Client
from .input_set import InputSet as InputSet
from .instance import Configuration as Configuration
from .instance import Instance as Instance
from .manifest import Manifest as Manifest
from .manifest import ManifestBuild as ManifestBuild
from .manifest import ManifestPython as ManifestPython
from .manifest import ManifestPythonModel as ManifestPythonModel
from .manifest import ManifestRuntime as ManifestRuntime
from .manifest import ManifestType as ManifestType
from .status import Status as Status
from .status import StatusV2 as StatusV2
from .version import Version as Version
from .version import VersionExecutable as VersionExecutable
from .version import VersionExecutableRequirements as VersionExecutableRequirements



from enum import Enum


class Status(str, Enum):
    """Status of a run. Deprecated: use StatusV2."""

    failed = "failed"
    """Run failed."""
    running = "running"
    """Run is running."""
    succeeded = "succeeded"
    """Run succeeded."""


class StatusV2(str, Enum):
    """Status of a run."""

    canceled = "canceled"
    """Run was canceled."""
    failed = "failed"
    """Run failed."""
    none = "none"
    """Run has no status."""
    queued = "queued"
    """Run is queued."""
    running = "running"
    """Run is running."""
    succeeded = "succeeded"
    """Run succeeded."""



from dataclasses import dataclass
from datetime import datetime

from nextmv.base_model import BaseModel
from nextmv.cloud.client import Client
from nextmv.cloud.status import Status, StatusV2


class QueuedRun(BaseModel):
    """A run that is pending to be executed in the account."""

    id: str
    """ID of the run."""
    user_email: str
    """Email of the user who created the run."""
    name: str
    """Name of the run."""
    description: str
    """Description of the run."""
    created_at: datetime
    """Creation date of the run."""
    application_id: str
    """ID of the application used for the run."""
    application_instance_id: str
    """ID of the application instance used for the run."""
    application_version_id: str
    """ID of the application version used for the run."""
    execution_class: str
    """Execution class used for the run."""
    status: Status
    """Deprecated: use status_v2."""
    status_v2: StatusV2
    """Status of the run."""


class Queue(BaseModel):
    """A queue is a list of runs that are pending to be executed, or currently
    being executed, in the account."""

    runs: list[QueuedRun]
    """List of runs in the queue."""


@dataclass
class Account:
    """The Nextmv Platform account."""

    client: Client
    """Client to use for interacting with the Nextmv Cloud API."""

    endpoint: str = "v1/account"
    """Base endpoint for the account."""

    def queue(self) -> Queue:
        """Get the queue of runs in the account.

        Returns:
            Queue: Queue of runs in the account.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
        """
        response = self.client.request(
            method="GET",
            endpoint=self.endpoint + "/queue",
        )

        return Queue.from_dict(response.json())



"""Configuration for a run."""

import argparse
import builtins
import copy
import os
from dataclasses import dataclass
from typing import Any, Optional

from nextmv.base_model import BaseModel


@dataclass
class Parameter:
    """
    Parameter that is used in a `Configuration`. When a parameter is required,
    it is a good practice to provide a default value for it. This is because
    the configuration will raise an error if a required parameter is not
    provided through a command-line argument, an environment variable or a
    default value.

    Parameters
    ----------
    name : str
        The name of the parameter.
    param_type : type
        The type of the parameter.
    default : Any, optional
        The default value of the parameter. Even though this is optional, it is
        recommended to provide a default value for all parameters.
    description : str, optional
        An optional description of the parameter. This is useful for generating
        help messages for the configuration.
    required : bool, optional
        Whether the parameter is required. If a parameter is required, it will
        be an error to not provide a value for it, either trough a command-line
        argument, an environment variable or a default value.
    choices : list[Optional[Any]], optional
        Limits values to a specific set of choices.
    """

    name: str
    """The name of the parameter."""
    param_type: type
    """The type of the parameter."""

    default: Optional[Any] = None
    """The default value of the parameter. Even though this is optional, it is
    recommended to provide a default value for all parameters."""
    description: Optional[str] = None
    """An optional description of the parameter. This is useful for generating
    help messages for the configuration."""
    required: bool = False
    """Whether the parameter is required. If a parameter is required, it will
    be an error to not provide a value for it, either trough a command-line
    argument, an environment variable or a default value."""
    choices: list[Optional[Any]] = None
    """Limits values to a specific set of choices."""

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Parameter":
        """
        Creates an instance of `Parameter` from a dictionary.

        Parameters
        ----------
        data : dict[str, Any]
            The dictionary representation of a parameter.

        Returns
        -------
        Parameter
            An instance of `Parameter`.
        """

        param_type_string = data["param_type"]
        param_type = getattr(builtins, param_type_string.split("'")[1])

        return Parameter(
            name=data["name"],
            param_type=param_type,
            default=data.get("default"),
            description=data.get("description"),
            required=data.get("required", False),
            choices=data.get("choices"),
        )

    def to_dict(self) -> dict[str, Any]:
        """
        Converts the parameter to a dict.

        Returns
        -------
        dict[str, Any]
            The parameter as a dict.
        """

        return {
            "name": self.name,
            "param_type": str(self.param_type),
            "default": self.default,
            "description": self.description,
            "required": self.required,
            "choices": self.choices,
        }


class Options:
    """
    Options for a run. To initialize options, pass in one or more `Parameter`
    objects. The options will look for the values of the given parameters in
    the following order: command-line arguments, environment variables, default
    values.

    Once the options are initialized, you can access the parameters as
    attributes of the `Options` object. For example, if you have a
    `Parameter` object with the name "duration", you can access it as
    `options.duration`.

    If a parameter is required and not provided through a command-line
    argument, an environment variable or a default value, an error will be
    raised.

    Options works as a Namespace, so you can assign new attributes to it. For
    example, you can do `options.foo = "bar"`.

    Options are parsed from the given sources when an attribute is accessed.
    Alternatively, you can call the `parse` method to parse the options
    manually. Options that are _not_ parsed may be merged with other unparsed
    options, by using the `merge` method. Once options are parsed, they cannot
    be merged with other options. After options are parsed, you may get the
    help message by running the script with the `-h/--help` flag.

    Parameters
    ----------
    *parameters : Parameter
        The parameters that are used in the options. At least one
        parameter is required.

    Examples
    --------
    >>> import nextmv
    >>>
    >>> options = nextmv.Options(
    ...     nextmv.Parameter("duration", str, "30s", description="solver duration", required=False),
    ...     nextmv.Parameter("threads", int, 4, description="computer threads", required=False),
    ... )
    >>>
    >>> print(options.duration, options.threads, options.to_dict())

    30s 4 {"duration": "30s", "threads": 4}

    Raises
    ------
    ValueError
        If a required parameter is not provided through a command-line
        argument, an environment variable or a default value.
    TypeError
        If a parameter is not a `Parameter` object.
    ValueError
        If an environment variable is not of the type of the corresponding
        parameter.
    """

    PARSED = False

    def __init__(self, *parameters: Parameter):
        """Initializes the options."""

        self.parameters = copy.deepcopy(parameters)

    def to_dict(self) -> dict[str, Any]:
        """
        Converts the options to a dict. As a side effect, this method parses
        the options if they have not been parsed yet. See the `parse` method
        for more information.

        Returns
        -------
        dict[str, Any]
            The options as a dict.
        """

        if not self.PARSED:
            self._parse()

        class model(BaseModel):
            config: dict[str, Any]

        self_dict = copy.deepcopy(self.__dict__)

        rm_keys = ["parameters", "PARSED"]
        for key in rm_keys:
            if key in self_dict:
                self_dict.pop(key)

        m = model.from_dict(data={"config": self_dict})

        return m.to_dict()["config"]

    def parameters_dict(self) -> list[dict[str, Any]]:
        """
        Converts the options to a list of dicts. Each dict is the dict
        representation of a `Parameter`.

        Returns
        -------
        list[dict[str, Any]]
            The list of dictionaries (parameter entries).
        """

        return [param.to_dict() for param in self.parameters]

    def parse(self):
        """
        Parses the options using command-line arguments, environment variables
        and default values, in that order. Under the hood, the `argparse`
        library is used. When command-line arguments are parsed, the help menu
        is created, thus parsing Options more than once may result in
        unexpected behavior.

        This method is called automatically when an attribute is accessed. If
        you want to parse the options manually, you can call this method.

        After Options have been parsed, they cannot be merged with other
        Options. If you need to merge Options, do so before parsing them.

        Example 1
        -------
        >>> import nextmv
        >>>
        >>> options = nextmv.Options(
        ...     nextmv.Parameter("duration", str, "30s", description="solver duration", required=False),
        ...     nextmv.Parameter("threads", int, 4, description="computer threads", required=False),
        ... )
        >>> options.parse() # Does not raise an exception.

        Example 2
        -------
        >>> import nextmv
        >>>
        >>> options = nextmv.Options(
        ...     nextmv.Parameter("duration", str, "30s", description="solver duration", required=False),
        ...     nextmv.Parameter("threads", int, 4, description="computer threads", required=False),
        ... )
        >>> print(options.duration) # Parses the options.
        >>> options.parse() # Raises an exception because the options have already been parsed.

        Raises
        ------
        RuntimeError
            If the options have already been parsed.
        ValueError
            If a required parameter is not provided through a command-line
            argument, an environment variable or a default value.
        TypeError
            If a parameter is not a `Parameter` object.
        ValueError
            If an environment variable is not of the type of the corresponding
            parameter.
        """

        if self.PARSED:
            raise RuntimeError("options have already been parsed")

        self._parse()

    def merge(self, new: "Options") -> "Options":
        """
        Merges the current options with the new options. This method cannot be
        used if any of the options have been parsed. When options are parsed,
        values are read from the command-line arguments, environment variables
        and default values. Merging options after parsing would result in
        unpredictable behavior.

        Parameters
        ----------
        new : Options
            The new options to merge.

        Raises
        ------
        RuntimeError
            If the current options have already been parsed.
        RuntimeError
            If the new options have already been parsed.

        Returns
        -------
        Options
            The merged options.
        """

        if self.PARSED:
            raise RuntimeError(
                "base options have already been parsed, cannot merge. See `Options.parse()` for more information."
            )

        if new.PARSED:
            raise RuntimeError(
                "new options have already been parsed, cannot merge. See `Options.parse()` for more information."
            )

        self.parameters += new.parameters

        self._parse()

        return self

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Options":
        """
        Creates an instance of `Options` from a dictionary. The dictionary
        should have the following structure:

        {
            "duration": "30",
            "threads": 4,
        }

        Parameters
        ----------
        data : dict[str, Any]
            The dictionary representation of the options.

        Returns
        -------
        Options
            An instance of `Options`.
        """

        parameters = []
        for key, value in data.items():
            parameter = Parameter(name=key, param_type=type(value), default=value)
            parameters.append(parameter)

        return cls(*parameters)

    @classmethod
    def from_parameters_dict(cls, parameters_dict: list[dict[str, Any]]) -> "Options":
        """
        Creates an instance of `Options` from parameters in dict form. Each
        entry is the dict representation of a `Parameter`.

        Parameters
        ----------
        data : list[dict[str, Any]]
            The list of dictionaries (parameter entries).

        Returns
        -------
        Options
            An instance of `Options`.
        """

        parameters = []
        for parameter_dict in parameters_dict:
            parameter = Parameter.from_dict(parameter_dict)
            parameters.append(parameter)

        return cls(*parameters)

    def __getattr__(self, name: str) -> Any:
        """
        Gets an attribute of the options. This is called when an attribute
        is accessed. It parses the options if they have not been parsed yet.
        """

        if not self.PARSED:
            self._parse()

        return super().__getattribute__(name)

    def _parse(self):  # noqa: C901
        """
        Parses the options using command-line arguments, environment variables
        and default values.

        Raises
        ------
        ValueError
            If a required parameter is not provided through a command-line
            argument, an environment variable or a default value.
        TypeError
            If a parameter is not a `Parameter` object.
        ValueError
            If an environment variable is not of the type of the corresponding
            parameter.
        """

        self.PARSED = True

        if not self.parameters:
            return

        parser = argparse.ArgumentParser(
            add_help=True,
            usage="%(prog)s [options]",
            description="Options for %(prog)s. Use command-line arguments (highest precedence) "
            + "or environment variables.",
            allow_abbrev=False,
        )
        params_by_field_name: dict[str, Parameter] = {}

        for p, param in enumerate(self.parameters):
            if not isinstance(param, Parameter):
                raise TypeError(f"expected a <Parameter> object, but got {type(param)} in index {p}")

            # See comment below about ipykernel adding a `-f` argument. We
            # restrict parameters from having the name 'f' or 'fff' for that
            # reason.
            if param.name == "f" or param.name == "fff":
                raise ValueError("parameter names 'f', 'fff' are reserved for internal use")

            if param.name == "PARSED":
                raise ValueError("parameter name 'PARSED' is reserved for internal use")

            # Remove any leading '-'. This is in line with argparse's behavior.
            param.name = param.name.lstrip("-")

            kwargs = {
                "type": param.param_type if param.param_type is not bool else str,
                "help": self._description(param),
            }

            if param.choices is not None:
                kwargs["choices"] = param.choices

            parser.add_argument(
                f"-{param.name}",
                f"--{param.name}",
                **kwargs,
            )

            # Store the parameter by its field name for easy access later. argparse
            # replaces '-' with '_', so we do the same here.
            params_by_field_name[param.name.replace("-", "_")] = param

        # The ipkyernel uses a `-f` argument by default that it passes to the
        # execution. We don’t want to ignore this argument because we get an
        # error. Fix source: https://stackoverflow.com/a/56349168
        parser.add_argument(
            "-f",
            "--f",
            "--fff",
            help=argparse.SUPPRESS,
            default="1",
        )
        args = parser.parse_args()

        for arg in vars(args):
            if arg == "fff" or arg == "f":
                continue

            param = params_by_field_name[arg]

            # First, attempt to set the value of a parameter from the
            # command-line args.
            arg_value = getattr(args, arg)
            if arg_value is not None:
                value = self._parameter_value(param, arg_value)
                setattr(self, arg, value)
                continue

            # Second, attempt to set the value of a parameter from the
            # environment variables.
            upper_name = arg.upper()
            env_value = os.getenv(upper_name)
            if env_value is not None:
                try:
                    typed_env_value = param.param_type(env_value) if param.param_type is not bool else env_value
                except ValueError:
                    raise ValueError(f'environment variable "{upper_name}" is not of type {param.param_type}') from None

                value = self._parameter_value(param, typed_env_value)
                setattr(self, arg, value)
                continue

            # Finally, attempt to set a default value. This is only allowed
            # for non-required parameters.
            if not param.required:
                setattr(self, arg, param.default)
                continue

            # At this point, the parameter is required and no value was
            # provided
            raise ValueError(
                f'parameter "{arg}" is required but not provided through: command-line args, env vars, or default value'
            )

    @staticmethod
    def _description(param: Parameter) -> str:
        """Returns a description for a parameter."""

        description = f"[env var: {param.name.upper()}]"

        if param.required:
            description += " (required)"

        if param.default is not None:
            description += f" (default: {param.default})"

        description += f" (type: {param.param_type.__name__})"

        if param.description is not None:
            description += f": {param.description}"

        return description

    @staticmethod
    def _parameter_value(parameter: Parameter, value: Any) -> Any:
        """Handles how the value of a parameter is extracted."""

        param_type = parameter.param_type
        if param_type is not bool:
            return value

        value = str(value).lower()

        if value in ("true", "1", "t", "y", "yes"):
            return True

        return False



import csv
import json
import os
import shutil
import unittest
from io import StringIO
from typing import Optional
from unittest.mock import patch

import nextmv


class TestOutput(unittest.TestCase):
    """Tests for the various classes for writing an output."""

    def test_local_writer_json_stdout_default(self):
        output = nextmv.Output(
            solution={"empanadas": "are_life"},
            statistics={"foo": "bar"},
        )
        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = {
                "solution": {"empanadas": "are_life"},
                "statistics": {"foo": "bar"},
                "options": {},
                "assets": [],
            }

            self.assertDictEqual(got, expected)

    def test_local_writer_json_stdout_default_dict_output(self):
        output = {
            "solution": {"empanadas": "are_life"},
            "statistics": {"foo": "bar"},
        }
        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = {
                "solution": {"empanadas": "are_life"},
                "statistics": {"foo": "bar"},
            }

            self.assertDictEqual(got, expected)

    def test_local_writer_json_stdout(self):
        output = nextmv.Output(
            output_format=nextmv.OutputFormat.JSON,
            solution={"empanadas": "are_life"},
            statistics={"foo": "bar"},
        )
        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = {
                "solution": {"empanadas": "are_life"},
                "statistics": {"foo": "bar"},
                "options": {},
                "assets": [],
            }

            self.assertDictEqual(got, expected)

    def test_local_writer_json_stdout_with_options(self):
        options = nextmv.Options()
        options.duration = 5
        options.solver = "highs"

        output = nextmv.Output(
            options=options,
            output_format=nextmv.OutputFormat.JSON,
            solution={"empanadas": "are_life"},
            statistics={"foo": "bar"},
        )
        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = {
                "options": {
                    "duration": 5,
                    "solver": "highs",
                },
                "solution": {"empanadas": "are_life"},
                "statistics": {"foo": "bar"},
                "assets": [],
            }

            self.assertDictEqual(got, expected)

    def test_local_writer_json_stdout_with_options_json(self):
        output = nextmv.Output(
            options={"duration": 5, "solver": "highs"},
            output_format=nextmv.OutputFormat.JSON,
            solution={"empanadas": "are_life"},
            statistics={"foo": "bar"},
        )
        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = {
                "options": {
                    "duration": 5,
                    "solver": "highs",
                },
                "solution": {"empanadas": "are_life"},
                "statistics": {"foo": "bar"},
                "assets": [],
            }

            self.assertDictEqual(got, expected)

    def test_local_writer_json_file(self):
        output = nextmv.Output(
            solution={"empanadas": "are_life"},
            statistics={"foo": "bar"},
        )
        output_writer = nextmv.LocalOutputWriter()

        with patch("builtins.open", create=True) as mock_open:
            output_writer.write(output, "output.json")

            handle = mock_open.return_value.__enter__.return_value
            handle.write.assert_called_once()

            got = json.loads(handle.write.call_args[0][0])
            expected = {
                "options": {},
                "solution": {"empanadas": "are_life"},
                "statistics": {"foo": "bar"},
                "assets": [],
            }

            self.assertDictEqual(got, expected)

    def test_local_writer_csvarchive_default_dir(self):
        """If the path for writing an output is not provided, the path `output`
        is used as the default directory."""
        self._test_local_writer_csvarchive(write_path="output", function_path="")

        # Should also work if not provided at all.
        self._test_local_writer_csvarchive(write_path="output", function_path=None)

    def test_local_writer_csvarchive_custom_dir(self):
        """Tests the flow of a CSV archive output writer but with a custom
        directory."""

        write_path = "KrAzYpAtH"
        self._test_local_writer_csvarchive(
            write_path=write_path,
            function_path=write_path,
        )

    def test_local_writer_csvarchive_wrong_path(self):
        output_writer = nextmv.LocalOutputWriter()
        output = nextmv.Output(
            output_format=nextmv.OutputFormat.CSV_ARCHIVE,
        )

        file_name = "a_file_should_not_be_specified.json"
        with open(file_name, "w") as file:
            file.write("")

        # Using a file that already exists should result in an error.
        with self.assertRaises(ValueError):
            # We patch stdout to avoid printing when executing the test.
            with patch("sys.stdout", new=StringIO()) as mock_stdout:
                output_writer.write(output, file_name, skip_stdout_reset=True)
                _ = mock_stdout.getvalue()

        os.remove(file_name)

        # However, using a file name as a directory should not result in an
        # error. It is kind of weird doing that, but to each their own.
        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, file_name, skip_stdout_reset=True)
            _ = mock_stdout.getvalue()

        # Removes the output directory after the test is executed.
        shutil.rmtree(file_name)

    def test_local_writer_csvarchive_dir_overwrite(self):
        output_dir = "empanadas_are_morally_superior_than_pizza"
        os.makedirs(output_dir, exist_ok=True)

        output_writer = nextmv.LocalOutputWriter()
        output = nextmv.Output(
            output_format=nextmv.OutputFormat.CSV_ARCHIVE,
        )

        # We patch stdout to avoid printing when executing the test.
        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, output_dir, skip_stdout_reset=True)
            _ = mock_stdout.getvalue()

        self.assertTrue(os.path.exists(output_dir))

        # Removes the output directory after the test is executed.
        shutil.rmtree(output_dir)

    def _test_local_writer_csvarchive(
        self,
        write_path: str,
        function_path: Optional[str] = None,
    ) -> None:
        """Auxiliary function that is used to test the flow of a CSV archive
        output output writer but with different directories."""

        options = nextmv.Options()
        options.parse()
        options.duration = 5
        options.solver = "highs"

        solution = {
            "empanadas": [
                {"are": 2.0, "life": 3.0},
                {"are": 5.0, "life": 6.0},
            ],
        }

        output = nextmv.Output(
            options=options,
            output_format=nextmv.OutputFormat.CSV_ARCHIVE,
            solution=solution,
            statistics={"foo": "bar"},
            csv_configurations={"quoting": csv.QUOTE_NONNUMERIC},
        )
        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, path=function_path, skip_stdout_reset=True)

            stdout_got = json.loads(mock_stdout.getvalue())
            stdout_expected = {
                "options": {
                    "duration": 5,
                    "solver": "highs",
                },
                "statistics": {"foo": "bar"},
                "assets": [],
            }

            self.assertDictEqual(stdout_got, stdout_expected)

        with open(f"{write_path}/empanadas.csv") as file:
            csv_got = file.read()

        csv_expected = '"are","life"\n2.0,3.0\n5.0,6.0\n'

        self.assertEqual(csv_got, csv_expected)

        self.assertTrue(os.path.exists(write_path))

        # Removes the output directory after the test is executed.
        shutil.rmtree(write_path)

    def test_local_write_bad_output_type(self):
        output = "I am clearly not an output object."
        with self.assertRaises(TypeError):
            nextmv.write_local(output)

    def test_local_write_passthrough_output(self):
        output = {
            "i_am": "a_crazy_object",
            "with": [
                {"nested": "values"},
                {"and": "more_craziness"},
            ],
        }

        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = output

            self.assertDictEqual(got, expected)

    def test_local_write_empty_output(self):
        output = nextmv.Output()

        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = expected = {
                "options": {},
                "solution": {},
                "statistics": {},
                "assets": [],
            }

            self.assertDictEqual(got, expected)

    def test_local_write_valid_assets_from_class(self):
        output = nextmv.Output(
            assets=[
                nextmv.Asset(
                    name="foo",
                    content={"foo": "bar"},
                    content_type="json",
                    description="A foo asset.",
                    visual=nextmv.Visual(
                        visual_schema=nextmv.VisualSchema.CHARTJS,
                        label="A chart",
                        visual_type="custom-tab",
                    ),
                ),
                nextmv.Asset(
                    name="bar",
                    content={"bar": "baz"},
                    content_type="json",
                    description="A bar asset.",
                ),
            ],
        )

        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = {
                "options": {},
                "solution": {},
                "statistics": {},
                "assets": [
                    {
                        "content": {"foo": "bar"},
                        "content_type": "json",
                        "description": "A foo asset.",
                        "name": "foo",
                        "visual": {
                            "label": "A chart",
                            "schema": "chartjs",
                            "type": "custom-tab",
                        },
                    },
                    {
                        "content": {"bar": "baz"},
                        "content_type": "json",
                        "description": "A bar asset.",
                        "name": "bar",
                    },
                ],
            }

            self.assertDictEqual(got, expected)

    def test_local_write_valid_assets_from_dict(self):
        assets = [
            {
                "name": "foo",
                "content": {"foo": "bar"},
                "content_type": "json",
                "description": "A foo asset.",
                "visual": {
                    "schema": "chartjs",
                    "label": "A chart",
                    "visual_type": "custom-tab",
                },
            },
            {
                "name": "bar",
                "content": {"bar": "baz"},
                "content_type": "json",
                "description": "A bar asset.",
            },
        ]
        output = nextmv.Output(assets=assets)

        output_writer = nextmv.LocalOutputWriter()

        with patch("sys.stdout", new=StringIO()) as mock_stdout:
            output_writer.write(output, skip_stdout_reset=True)

            got = json.loads(mock_stdout.getvalue())
            expected = {
                "options": {},
                "solution": {},
                "statistics": {},
                "assets": assets,
            }

            self.assertDictEqual(got, expected)



"""This module contains definitions for acceptance tests."""

from datetime import datetime
from enum import Enum
from typing import Optional

from nextmv.base_model import BaseModel


class MetricType(str, Enum):
    """Type of metric when doing a comparison."""

    direct_comparison = "direct-comparison"
    """Direct comparison metric type."""


class StatisticType(str, Enum):
    """
    Type of statistical process for collapsing multiple values of a metric
    (from multiple runs) into a single value.
    """

    min = "min"
    """Minimum value."""
    max = "max"
    """Maximum value."""
    mean = "mean"
    """Mean value."""
    std = "std"
    """Standard deviation."""
    shifted_geometric_mean = "shifted_geometric_mean"
    """Shifted geometric mean."""
    p01 = "p01"
    """1st percentile."""
    p05 = "p05"
    """5th percentile."""
    p10 = "p10"
    """10th percentile."""
    p25 = "p25"
    """25th percentile."""
    p50 = "p50"
    """50th percentile."""
    p75 = "p75"
    """75th percentile."""
    p90 = "p90"
    """90th percentile."""
    p95 = "p95"
    """95th percentile."""
    p99 = "p99"
    """99th percentile."""


class Comparison(str, Enum):
    """Comparison to use for two metrics."""

    equal_to = "eq"
    """Equal to metric type."""
    greater_than = "gt"
    """Greater than metric type."""
    greater_than_or_equal_to = "ge"
    """Greater than or equal to metric type."""
    less_than = "lt"
    """Less than metric type."""
    less_than_or_equal_to = "le"
    """Less than or equal to metric type."""
    not_equal_to = "ne"
    """Not equal to metric type."""


class ToleranceType(str, Enum):
    """Type of tolerance used for a metric."""

    undefined = ""
    """Undefined tolerance type."""
    absolute = "absolute"
    """Absolute tolerance type."""
    relative = "relative"
    """Relative tolerance type."""


class ExperimentStatus(str, Enum):
    """Status of an acceptance test."""

    started = "started"
    """The experiment has started."""
    completed = "completed"
    """The experiment was completed."""
    failed = "failed"
    """The experiment failed."""
    draft = "draft"
    """The experiment is a draft."""
    canceled = "canceled"
    """The experiment was canceled."""
    unknown = "unknown"
    """The experiment status is unknown."""


class MetricTolerance(BaseModel):
    """Tolerance used for a metric."""

    type: ToleranceType
    """Type of tolerance."""
    value: float
    """Value of the tolerance."""


class MetricParams(BaseModel):
    """Parameters of an acceptance test."""

    operator: Comparison
    """Operator used to compare two metrics."""
    tolerance: MetricTolerance
    """Tolerance used for the comparison."""


class Metric(BaseModel):
    """A metric is a key performance indicator that is used to evaluate the
    performance of a test."""

    field: str
    """Field of the metric."""
    metric_type: MetricType
    """Type of the metric."""
    params: MetricParams
    """Parameters of the metric."""
    statistic: StatisticType
    """
    Type of statistical process for collapsing multiple values of a metric
    (from multiple runs) into a single value.
    """


class ComparisonInstance(BaseModel):
    """An app instance used for a comparison."""

    instance_id: str
    """ID of the instance."""
    version_id: str
    """ID of the version."""


class DistributionSummaryStatistics(BaseModel):
    """Statistics of a distribution summary."""

    min: float
    """Minimum value."""
    max: float
    """Maximum value."""
    count: int
    """Count of runs."""
    mean: float
    """Mean value."""
    std: float
    """Standard deviation."""
    shifted_geometric_mean: float
    """Shifted geometric mean."""
    shift_parameter: float
    """Shift parameter of the geometric mean."""


class DistributionPercentiles(BaseModel):
    """Percentiles of a distribution."""

    p01: float
    """1st percentile."""
    p05: float
    """5th percentile."""
    p10: float
    """10th percentile."""
    p25: float
    """25th percentile."""
    p50: float
    """50th percentile."""
    p75: float
    """75th percentile."""
    p90: float
    """90th percentile."""
    p95: float
    """95th percentile."""
    p99: float
    """99th percentile."""


class ResultStatistics(BaseModel):
    """Statistics of a metric result."""

    instance_id: str
    """ID of the instance."""
    version_id: str
    """ID of the version."""
    number_of_runs_total: int
    """Number of runs."""
    distribution_summary_statistics: DistributionSummaryStatistics
    """Distribution summary statistics."""
    distribution_percentiles: DistributionPercentiles
    """Distribution percentiles."""


class MetricStatistics(BaseModel):
    """Statistics of a metric."""

    control: ResultStatistics
    """Control statistics."""
    candidate: ResultStatistics
    """Candidate statistics."""


class MetricResult(BaseModel):
    """Result of a metric."""

    metric: Metric
    """Metric of the result."""
    statistics: MetricStatistics
    """Statistics of the metric."""
    passed: bool
    """Whether the candidate passed for the metric (or not)."""


class AcceptanceTestResults(BaseModel):
    """Results of an acceptance test."""

    passed: bool
    """Whether the acceptance test passed (or not)."""
    metric_results: Optional[list[MetricResult]] = None
    """Results of the metrics."""
    error: Optional[str] = None
    """Error message if the acceptance test failed."""


class AcceptanceTest(BaseModel):
    """An acceptance test gives a go/no-go decision criteria for a set of
    metrics. It relies on a batch experiment."""

    id: str
    """ID of the acceptance test."""
    name: str
    """Name of the acceptance test."""
    description: str
    """Description of the acceptance test."""
    app_id: str
    """ID of the app that owns the acceptance test."""
    experiment_id: str
    """ID of the batch experiment underlying in the acceptance test."""
    control: ComparisonInstance
    """Control instance of the acceptance test."""
    candidate: ComparisonInstance
    """Candidate instance of the acceptance test."""
    metrics: list[Metric]
    """Metrics of the acceptance test."""
    created_at: datetime
    """Creation date of the acceptance test."""
    updated_at: datetime
    """Last update date of the acceptance test."""
    status: Optional[ExperimentStatus] = ExperimentStatus.unknown
    """Status of the acceptance test."""
    results: Optional[AcceptanceTestResults] = None
    """Results of the acceptance test."""



import unittest

from nextmv.cloud.manifest import Manifest, ManifestPython, ManifestRuntime, ManifestType
from nextmv.model import ModelConfiguration
from nextmv.options import Options, Parameter


class TestManifest(unittest.TestCase):
    def test_from_model_configuration(self):
        options = Options(
            Parameter("param1", str, ""),
            Parameter("param2", str, ""),
        )
        model_configuration = ModelConfiguration(
            name="super_cool_model",
            requirements=[
                "one_requirement",
                "another_requirement",
            ],
            options=options,
        )
        manifest = Manifest.from_model_configuration(model_configuration)

        self.assertListEqual(
            manifest.files,
            ["main.py", f"{model_configuration.name}/**"],
        )
        self.assertEqual(manifest.runtime, ManifestRuntime.PYTHON)
        self.assertEqual(manifest.type, ManifestType.PYTHON)

        manifest_python = ManifestPython.from_dict(
            {
                "pip-requirements": "model_requirements.txt",
                "model": {
                    "name": model_configuration.name,
                    "options": model_configuration.options.parameters_dict(),
                },
            }
        )
        self.assertEqual(manifest.python, manifest_python)



"""This module contains definitions for batch experiments."""

from datetime import datetime
from typing import Any, Optional

from nextmv.base_model import BaseModel


class BatchExperimentInformation(BaseModel):
    """Information about a batch experiment. This serves as a base for all the
    other batch experiment models."""

    name: str
    """Name of the batch experiment."""
    input_set_id: str
    """ID of the input set used for the experiment."""
    instance_ids: list[str]
    """List of instance IDs used for the experiment."""

    description: Optional[str] = None
    """Description of the batch experiment."""
    id: Optional[str] = None
    """ID of the batch experiment."""


class BatchExperiment(BatchExperimentInformation):
    """A batch experiment compares two or more instances by executing all the
    inputs contained in the input set."""

    created_at: datetime
    """Creation date of the batch experiment."""
    status: str
    """Status of the batch experiment."""

    grouped_distributional_summaries: Optional[list[dict[str, Any]]] = None
    """Grouped distributional summaries of the batch experiment."""
    option_sets: Optional[dict[str, dict[str, str]]] = None
    """Option sets used for the experiment."""


class BatchExperimentRun(BaseModel):
    """A batch experiment run is a single execution of a batch experiment."""

    option_set: str
    """Option set used for the experiment."""
    input_id: str
    """ID of the input used for the experiment."""

    instance_id: Optional[str] = None
    """ID of the instance used for the experiment."""
    version_id: Optional[str] = None
    """ID of the version used for the experiment."""

    def __post_init__(self):
        """Logic to run after the class is initialized."""

        if self.instance_id is None and self.version_id is None:
            raise ValueError("either instance_id or version_id must be set")


class BatchExperimentMetadata(BatchExperimentInformation):
    """Metadata of a batch experiment."""

    status: str
    """Status of the batch experiment."""
    created_at: datetime
    """Creation date of the batch experiment."""
    number_of_runs: int
    """Number of runs in the batch experiment."""



"""This module contains definitions for input sets."""

from datetime import datetime

from nextmv.base_model import BaseModel


class InputSet(BaseModel):
    """An input set is the collection of inputs from the associated runs."""

    app_id: str
    """ID of the application that the input set belongs to."""
    created_at: datetime
    """Creation time of the input set."""
    description: str
    """Description of the input set."""
    id: str
    """ID of the input set."""
    input_ids: list[str]
    """IDs of the inputs in the input set."""
    name: str
    """Name of the input set."""
    updated_at: datetime
    """Last update time of the input set."""



import logging
import os
import shutil
from dataclasses import dataclass
from typing import Any, Optional

from nextmv.input import Input
from nextmv.logger import log
from nextmv.options import Options
from nextmv.output import Output

# When working with the `Model`, we expect to be working in a notebook
# environment, and not interact with the local filesystem a lot. We use the
# `ModelConfiguration` to specify the dependencies that the `Model` requires.
# To work with the "push" logic of uploading an app to Nextmv Cloud, we need a
# requirement file that we use to gather dependencies, install them, and bundle
# them in the app. This file is used as a placeholder for the dependencies that
# the model requires and that we install and bundle with the app.
_REQUIREMENTS_FILE = "model_requirements.txt"

# When working in a notebook environment, we don’t really create a `main.py`
# file with the main entrypoint of the program. Because the logic is mostly
# encoded inside the `Model` class, we need to create a `main.py` file that we
# can run in Nextmv Cloud. This file is used as that entrypoint.
_ENTRYPOINT_FILE = "__entrypoint__.py"

_MLFLOW_DEPENDENCY = "mlflow>=2.18.0"


@dataclass
class ModelConfiguration:
    """
    ModelConfiguration is a class that holds the configuration for a
    model. It is used to define how a Python model is encoded and loaded.

    The `name` is required, and should be a personalized name for the model.

    You may specify the `requirements` that your decision model requires. This
    is done by passing a list of requirements, as if they were lines in a
    `requirements.txt` file. An example of this is `["nextmv==0.1.0"]`.

    Lastly, if your decision model requires options, you may specify them by
    passing an instance of `Options`.
    """

    name: str
    """The name of the decision model."""

    requirements: Optional[list[str]] = None
    """A list of Python dependencies that the decision model requires."""
    options: Optional[Options] = None
    """Options that the decision model requires."""


class Model:
    """
    Model is the base class for defining a decision model that runs in Nextmv
    Cloud. You must create a subclass of this class and implement the `solve`
    method. The `solve` method is the main entry point of your model and should
    return an output with a solution (decision).

    Example
    -------
    ```python
    import nextroute

    import nextmv


    # Define the model that makes decisions. This model uses the Nextroute library
    # to solve a routing problem.
    class DecisionModel(nextmv.Model):
        def solve(self, input: nextmv.Input) -> nextmv.Output:
            nextroute_input = nextroute.schema.Input.from_dict(input.data)
            nextroute_options = nextroute.Options.extract_from_dict(input.options.to_dict())
            nextroute_output = nextroute.solve(nextroute_input, nextroute_options)

            return nextmv.Output(
                options=input.options,
                solution=nextroute_output.solutions[0].to_dict(),
                statistics=nextroute_output.statistics.to_dict(),
            )
        ```
    """

    def solve(self, input: Input) -> Output:
        """
        The `solve` method is the main entry point of your model. You must
        implement this method yourself. It receives a `nextmv.Input` and should
        process it to produce a `nextmv.Output`, which is the solution to the
        decision model/problem.

        Parameters
        ----------
        input : Input
            The input data that the model will use to make a decision.

        Returns
        -------
        Output
            The output of the model, which is the solution to the decision
            model/problem.
        """

        raise NotImplementedError

    def save(self, model_dir: str, configuration: ModelConfiguration) -> None:
        """
        Save the model to the local filesystem, in the location given by `dir`.
        The model is saved according to the configuration provided, which is of
        type `ModelConfiguration`.

        Parameters
        ----------
        dir : str
            The directory where the model will be saved.
        configuration : ModelConfiguration
            The configuration of the model, which defines how the model is
            saved and loaded.
        """

        # mlflow is a big package. We don’t want to make it a dependency of
        # `nextmv` because it is not always needed. We only need it if we are
        # working with the "app from model" logic, which involves working with
        # this `Model` class.
        try:
            import mlflow as mlflow
        except ImportError as e:
            raise ImportError(
                "mlflow is not installed. Please install optional dependencies with `pip install nextmv[all]`"
            ) from e

        finally:
            from mlflow.models import infer_signature
            from mlflow.pyfunc import PythonModel, save_model

        class MLFlowModel(PythonModel):
            """
            The `MLFlowModel` class exists as a transient class to translate a
            Nextmv `DecisionModel` into an `mlflow.pyfunc.PythonModel`. This
            class must comply with the inference API of mlflow, which is why it
            has a `predict` method. The translation happens by having this
            `predict` method call the user-defined `solve` method of the
            `DecisionModel`.
            """

            def predict(mlflow_self, context, model_input, params=None) -> Any:
                """
                The predict method allows us to work with mlflow’s [python_function]
                model flavor. Warning: This method should not be used or overridden
                directly. Instead, you should implement the `solve` method.

                [python_function]: https://mlflow.org/docs/latest/python_api/mlflow.pyfunc.html
                """

                return self.solve(model_input)

        # Some annoying logging from mlflow must be disabled.
        logging.disable(logging.CRITICAL)

        _cleanup_python_model(model_dir, configuration, verbose=False)

        signature = None
        if configuration.options is not None:
            options_dict = configuration.options.to_dict()
            signature = infer_signature(
                params=options_dict,
            )

        # We use mlflow to save the model to the local filesystem, to be able to
        # load it later on.
        model_path = os.path.join(model_dir, configuration.name)
        save_model(
            path=model_path,  # Customize the name of the model location.
            infer_code_paths=True,  # Makes the imports portable.
            python_model=MLFlowModel(),
            signature=signature,  # Allows us to work with our own `Options` class.
        )

        logging.disable(logging.NOTSET)

        # Create an auxiliary requirements file with the model dependencies.
        requirements_file = os.path.join(model_dir, _REQUIREMENTS_FILE)
        with open(requirements_file, "w") as file:
            file.write(f"{_MLFLOW_DEPENDENCY}\n")
            reqs = configuration.requirements
            if reqs is not None:
                for req in reqs:
                    file.write(f"{req}\n")

        # Adds the main.py file to the app_dir by coping the `entrypoint.py` file
        # which is one level up from this file.
        entrypoint_file = os.path.join(os.path.dirname(__file__), _ENTRYPOINT_FILE)
        shutil.copy2(entrypoint_file, os.path.join(model_dir, "main.py"))


def _cleanup_python_model(
    model_dir: str,
    model_configuration: Optional[ModelConfiguration] = None,
    verbose: bool = False,
) -> None:
    """Cleans up the Python-specific model packaging logic."""

    if model_configuration is None:
        return

    model_path = os.path.join(model_dir, model_configuration.name)
    if os.path.exists(model_path):
        shutil.rmtree(model_path)

    mlruns_path = os.path.join(model_dir, "mlruns")
    if os.path.exists(mlruns_path):
        shutil.rmtree(mlruns_path)

    requirements_file = os.path.join(model_dir, _REQUIREMENTS_FILE)
    if os.path.exists(requirements_file):
        os.remove(requirements_file)

    main_file = os.path.join(model_dir, "main.py")
    if os.path.exists(main_file):
        os.remove(main_file)

    if verbose:
        log("🧹 Cleaned up Python model artifacts.")



"""
When working in a notebook environment, we don’t really create a `main.py` file
with the main entrypoint of the program. Because the logic is mostly encoded
inside the `Model` class, we need to create a `main.py` file that we can run in
Nextmv Cloud. This file is used as that entrypoint. It is not intended for a
human to use it during local development. It is the standard way in which a
`nextmv.Model` is loaded by using mlflow.
"""

from mlflow.pyfunc import load_model

from nextmv.cloud.manifest import Manifest
from nextmv.input import load_local
from nextmv.options import Options
from nextmv.output import write_local


def main() -> None:
    """Entry point for the program."""

    manifest = Manifest.from_yaml(".")

    # Load the options from the manifest.
    options = None
    parameters_dict = manifest.python.model.options
    if parameters_dict is not None:
        options = Options.from_parameters_dict(parameters_dict)

    # Load the model.
    loaded_model = load_model(
        model_uri=manifest.python.model.name,
        suppress_warnings=True,
    )

    # Load the input and solve the model by using mlflow’s inference API.
    input = load_local(options=options)
    output = loaded_model.predict(input)

    # Write the output.
    write_local(output)


if __name__ == "__main__":
    main()



"""Module with the client class."""

import json
import os
from dataclasses import dataclass, field
from typing import IO, Any, Optional, Union
from urllib.parse import urljoin

import requests
import yaml
from requests.adapters import HTTPAdapter, Retry

_MAX_LAMBDA_PAYLOAD_SIZE: int = 500 * 1024 * 1024
"""Maximum size of the payload handled by the Nextmv Cloud API."""


@dataclass
class Client:
    """
    Client that interacts directly with the Nextmv Cloud API. The API key will
    be searched, in order of precedence, in: the api_key arg in the
    constructor, the NEXTMV_API_KEY environment variable, the
    ~/.nextmv/config.yaml file used by the Nextmv CLI.
    """

    api_key: Optional[str] = None
    """API key to use for authenticating with the Nextmv Cloud API. If not
    provided, the client will look for the NEXTMV_API_KEY environment
    variable."""
    allowed_methods: list[str] = field(
        default_factory=lambda: ["GET", "POST", "PUT", "DELETE"],
    )
    """Allowed HTTP methods to use for retries in requests to the Nextmv Cloud
    API."""
    backoff_factor: float = 1
    """Exponential backoff factor to use for requests to the Nextmv Cloud
    API."""
    backoff_jitter: float = 0.1
    """Jitter to use for requests to the Nextmv Cloud API when backing off."""
    backoff_max: float = 60
    """Maximum backoff time to use for requests to the Nextmv Cloud API, in
    seconds."""
    configuration_file: str = "~/.nextmv/config.yaml"
    """Path to the configuration file used by the Nextmv CLI."""
    headers: Optional[dict[str, str]] = None
    """Headers to use for requests to the Nextmv Cloud API."""
    max_retries: int = 10
    """Maximum number of retries to use for requests to the Nextmv Cloud
    API."""
    status_forcelist: list[int] = field(
        default_factory=lambda: [429, 500, 502, 503, 504, 507, 509],
    )
    """Status codes to retry for requests to the Nextmv Cloud API."""
    timeout: float = 20
    """Timeout to use for requests to the Nextmv Cloud API."""
    url: str = "https://api.cloud.nextmv.io"
    """URL of the Nextmv Cloud API."""

    def __post_init__(self):
        """Logic to run after the class is initialized."""

        if self.api_key is not None and self.api_key != "":
            self._set_headers_api_key(self.api_key)
            return

        if self.api_key == "":
            raise ValueError("api_key cannot be empty")

        api_key_env = os.getenv("NEXTMV_API_KEY")
        if api_key_env is not None:
            self.api_key = api_key_env
            self._set_headers_api_key(api_key_env)
            return

        config_path = os.path.expanduser(self.configuration_file)
        if not os.path.exists(config_path):
            raise ValueError(
                "no API key set in constructor or NEXTMV_API_KEY env var, and ~/.nextmv/config.yaml does not exist"
            )

        with open(config_path) as f:
            config = yaml.safe_load(f)

        profile = os.getenv("NEXTMV_PROFILE")
        parent = config
        if profile is not None:
            parent = config.get(profile)
            if parent is None:
                raise ValueError(f"profile {profile} set via NEXTMV_PROFILE but not found in ~/.nextmv/config.yaml")

        api_key = parent.get("apikey")
        if api_key is None:
            raise ValueError("no apiKey found in ~/.nextmv/config.yaml")
        self.api_key = api_key

        endpoint = parent.get("endpoint")
        if endpoint is not None:
            self.url = f"https://{endpoint}"

        self._set_headers_api_key(api_key)

    def request(
        self,
        method: str,
        endpoint: str,
        data: Optional[Any] = None,
        headers: Optional[dict[str, str]] = None,
        payload: Optional[dict[str, Any]] = None,
        query_params: Optional[dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Method to make a request to the Nextmv Cloud API.

        Args:
            method: HTTP method to use. Valid methods include: GET, POST.
            endpoint: Endpoint to send the request to.
            data: Data to send with the request.
            headers: Headers to send with the request.
            payload: Payload to send with the request. Prefer using this over
                data.
            query_params: Query parameters to send with the request.

        Returns:
            Response from the Nextmv Cloud API.

        Raises:
            requests.HTTPError: If the response status code is not 2xx.
            ValueError: If both data and payload are provided.
            ValueError: If the payload size exceeds the maximum allowed size.
            ValueError: If the data size exceeds the maximum allowed size.
        """

        if payload is not None and data is not None:
            raise ValueError("cannot use both data and payload")

        if payload is not None and get_size(payload) > _MAX_LAMBDA_PAYLOAD_SIZE:
            raise ValueError(
                f"payload size of {get_size(payload)} bytes exceeds the maximum "
                f"allowed size of {_MAX_LAMBDA_PAYLOAD_SIZE} bytes"
            )

        if data is not None and get_size(data) > _MAX_LAMBDA_PAYLOAD_SIZE:
            raise ValueError(
                f"data size of {get_size(data)} bytes exceeds the maximum "
                f"allowed size of {_MAX_LAMBDA_PAYLOAD_SIZE} bytes"
            )

        session = requests.Session()
        retries = Retry(
            total=self.max_retries,
            backoff_factor=self.backoff_factor,
            backoff_jitter=self.backoff_jitter,
            backoff_max=self.backoff_max,
            status_forcelist=self.status_forcelist,
            allowed_methods=self.allowed_methods,
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount("https://", adapter)

        kwargs = {
            "url": urljoin(self.url, endpoint),
            "timeout": self.timeout,
        }
        kwargs["headers"] = headers if headers is not None else self.headers
        if data is not None:
            kwargs["data"] = data
        if payload is not None:
            kwargs["json"] = payload
        if query_params is not None:
            kwargs["params"] = query_params

        response = session.request(method=method, **kwargs)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            raise requests.HTTPError(
                f"request to {endpoint} failed with status code {response.status_code} and message: {response.text}"
            ) from e

        return response

    def upload_to_presigned_url(
        self,
        data: Union[dict[str, Any], str],
        url: str,
    ) -> None:
        """
        Method to upload data to a presigned URL of the Nextmv Cloud API.
        Args:
            data: data to upload.
            url: URL to upload the data to.
        """

        upload_data = None
        if isinstance(data, dict):
            upload_data = json.dumps(data, separators=(",", ":"))
        elif isinstance(data, str):
            upload_data = data
        else:
            raise ValueError("data must be a dictionary or a string")

        session = requests.Session()
        retries = Retry(
            total=self.max_retries,
            backoff_factor=self.backoff_factor,
            backoff_jitter=self.backoff_jitter,
            backoff_max=self.backoff_max,
            status_forcelist=self.status_forcelist,
            allowed_methods=self.allowed_methods,
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount("https://", adapter)
        kwargs = {
            "url": url,
            "timeout": self.timeout,
            "data": upload_data,
        }

        response = session.put(**kwargs)

        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            raise requests.HTTPError(
                f"upload to presigned URL {url} failed with "
                + f"status code {response.status_code} and message: {response.text}"
            ) from e

    def _set_headers_api_key(self, api_key: str) -> None:
        """Sets the API key to use for requests to the Nextmv Cloud API."""

        self.headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }


def get_size(obj: Union[dict[str, Any], IO[bytes]]) -> int:
    """Finds the size of an object in bytes."""

    if isinstance(obj, dict):
        obj_str = json.dumps(obj, separators=(",", ":"))
        return len(obj_str.encode("utf-8"))

    elif hasattr(obj, "read"):
        obj.seek(0, 2)  # Move the cursor to the end of the file
        size = obj.tell()
        obj.seek(0)  # Reset the cursor to the beginning of the file
        return size

    elif isinstance(obj, str):
        return len(obj.encode("utf-8"))

    else:
        raise TypeError("Unsupported type. Only dictionaries and file objects are supported.")



import csv
import os
import shutil
import unittest
from io import StringIO
from typing import Optional
from unittest.mock import patch

import nextmv


class TestInput(unittest.TestCase):
    """
    Tests for the various classes for loading an input.
    """

    def test_local_loader_json_stdin(self):
        sample_input = '{"empanadas": "are_life"}\n'
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load()

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.JSON)
        self.assertEqual(input_data.data, {"empanadas": "are_life"})
        self.assertIsNone(input_data.options)

    def test_local_loader_text_stdin(self):
        sample_input = "empanadas are life\n"
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load(input_format=nextmv.InputFormat.TEXT)

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.TEXT)
        self.assertEqual(input_data.data, "empanadas are life")
        self.assertIsNone(input_data.options)

    def test_local_loader_csv_stdin(self):
        sample_input = '"empanadas","are","life"\n1,2,3\n4,5,6'
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load(
                input_format=nextmv.InputFormat.CSV,
                csv_configurations={
                    "quoting": csv.QUOTE_NONNUMERIC,
                },
            )

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.CSV)
        self.assertEqual(
            list(input_data.data),
            [
                {"empanadas": 1.0, "are": 2.0, "life": 3.0},
                {"empanadas": 4.0, "are": 5.0, "life": 6.0},
            ],
        )
        self.assertIsNone(input_data.options)

    def test_local_loader_with_options(self):
        sample_input = '{"empanadas": "are_life"}\n'
        options = nextmv.Options(nextmv.Parameter("foo", str, default="bar", required=False))
        input_loader = nextmv.LocalInputLoader()

        with patch("sys.stdin", new=StringIO(sample_input)):
            input_data = input_loader.load(options=options)

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.JSON)
        self.assertEqual(input_data.data, {"empanadas": "are_life"})
        self.assertIsNotNone(input_data.options)
        self.assertDictEqual(input_data.options.to_dict(), options.to_dict())

    def test_local_loader_json_file(self):
        sample_input = '{"empanadas": "are_life"}'
        input_loader = nextmv.LocalInputLoader()

        with patch("builtins.open", return_value=StringIO(sample_input)):
            input_data = input_loader.load(path="input.json")

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.JSON)
        self.assertEqual(input_data.data, {"empanadas": "are_life"})
        self.assertIsNone(input_data.options)

    def test_local_loader_text_file(self):
        sample_input = "empanadas are life"
        input_loader = nextmv.LocalInputLoader()

        with patch("builtins.open", return_value=StringIO(sample_input)):
            input_data = input_loader.load(input_format=nextmv.InputFormat.TEXT, path="input.txt")

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.TEXT)
        self.assertEqual(input_data.data, "empanadas are life")
        self.assertIsNone(input_data.options)

    def test_local_loader_csv_file(self):
        sample_input = '"empanadas","are","life"\n1,2,3\n4,5,6'
        input_loader = nextmv.LocalInputLoader()

        with patch("builtins.open", return_value=StringIO(sample_input)):
            input_data = input_loader.load(
                input_format=nextmv.InputFormat.CSV,
                path="input.csv",
                csv_configurations={
                    "quoting": csv.QUOTE_NONNUMERIC,
                },
            )

        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.CSV)
        self.assertEqual(
            list(input_data.data),
            [
                {"empanadas": 1.0, "are": 2.0, "life": 3.0},
                {"empanadas": 4.0, "are": 5.0, "life": 6.0},
            ],
        )
        self.assertIsNone(input_data.options)

    def test_local_loader_csv_archive_default_dir(self):
        """If the path for loading the input is not provided, the path `input`
        is used for the directory."""
        self._test_local_loader_csv_archive(test_dir="input", load_path="")

        # Should also work if not provided at all.
        self._test_local_loader_csv_archive(test_dir="input", load_path=None)

    def test_local_loader_csv_archive_custom_dir(self):
        """If the path for loading the input is provided, the path is used for
        the directory."""
        self._test_local_loader_csv_archive(test_dir="custom_dir", load_path="custom_dir")

    def _test_local_loader_csv_archive(
        self,
        test_dir: str,
        load_path: Optional[str] = None,
    ):
        """This is an auxiliary function that is used to test the flow of the
        CSV archive input loader but with different directories."""

        # Create the directory if it doesn't exist
        os.makedirs(test_dir, exist_ok=True)

        input_loader = nextmv.LocalInputLoader()

        # Write sample CSV files
        sample_input_1 = '"empanadas","are","life"\n1,2,3\n4,5,6'
        sample_input_2 = '"or","are","tacos"\n7,8,9\n10,11,12'
        with open(f"{test_dir}/empanada_declaration_archive.csv", "w") as file_1:
            file_1.write(sample_input_1)
        with open(f"{test_dir}/taco_declaration_archive.csv", "w") as file_2:
            file_2.write(sample_input_2)

        # Load the CSV archive input
        input_data = input_loader.load(
            nextmv.InputFormat.CSV_ARCHIVE,
            path=load_path,
            csv_configurations={
                "quoting": csv.QUOTE_NONNUMERIC,
            },
        )

        # Do the checks
        self.assertIsInstance(input_data, nextmv.Input)
        self.assertEqual(input_data.input_format, nextmv.InputFormat.CSV_ARCHIVE)
        self.assertIsNone(input_data.options)

        self.assertIn(
            "empanada_declaration_archive",
            list(input_data.data.keys()),
        )
        self.assertIn(
            "taco_declaration_archive",
            list(input_data.data.keys()),
        )
        self.assertEqual(
            list(input_data.data["empanada_declaration_archive"]),
            [
                {"empanadas": 1.0, "are": 2.0, "life": 3.0},
                {"empanadas": 4.0, "are": 5.0, "life": 6.0},
            ],
        )
        self.assertEqual(
            list(input_data.data["taco_declaration_archive"]),
            [
                {"or": 7.0, "are": 8.0, "tacos": 9.0},
                {"or": 10.0, "are": 11.0, "tacos": 12.0},
            ],
        )

        # Remove the directory.
        shutil.rmtree(test_dir)


if __name__ == "__main__":
    unittest.main()



"""This module contains definitions for instances."""

from datetime import datetime
from typing import Optional

from nextmv.base_model import BaseModel


class Configuration(BaseModel):
    """Configuration for an instance."""

    execution_class: Optional[str] = None
    """Execution class for the instance."""
    options: Optional[dict] = None
    """Options of the app that the instance uses."""


class Instance(BaseModel):
    """A instance of an application tied to a version and further configuration."""

    id: str
    """ID of the instance."""
    application_id: str
    """ID of the application that this is an instance of."""
    version_id: str
    """ID of the version that this instance is uses."""
    name: str
    """Name of the instance."""
    description: str
    """Description of the instance."""
    configuration: Configuration
    """Configuration for the instance."""
    locked: bool
    """Whether the instance is locked."""
    created_at: datetime
    """Creation time of the instance."""
    updated_at: datetime
    """Last update time of the instance."""



"""Nextmv Python SDK."""

from .__about__ import __version__
from .base_model import from_dict as from_dict
from .input import Input as Input
from .input import InputFormat as InputFormat
from .input import InputLoader as InputLoader
from .input import LocalInputLoader as LocalInputLoader
from .input import load_local as load_local
from .logger import log as log
from .logger import redirect_stdout as redirect_stdout
from .logger import reset_stdout as reset_stdout
from .model import Model as Model
from .model import ModelConfiguration as ModelConfiguration
from .options import Options as Options
from .options import Parameter as Parameter
from .output import Asset as Asset
from .output import DataPoint as DataPoint
from .output import LocalOutputWriter as LocalOutputWriter
from .output import Output as Output
from .output import OutputFormat as OutputFormat
from .output import OutputWriter as OutputWriter
from .output import ResultStatistics as ResultStatistics
from .output import RunStatistics as RunStatistics
from .output import Series as Series
from .output import SeriesData as SeriesData
from .output import Statistics as Statistics
from .output import Visual as Visual
from .output import VisualSchema as VisualSchema
from .output import write_local as write_local

VERSION = __version__
"""The version of the Nextmv Python SDK."""



"""This module contains definitions for versions."""

from datetime import datetime

from nextmv.base_model import BaseModel


class VersionExecutableRequirements(BaseModel):
    """Requirements for a version executable."""

    executable_type: str
    """Type of the executable."""
    runtime: str
    """Runtime for the executable."""


class VersionExecutable(BaseModel):
    """Executable for a version."""

    id: str
    """ID of the version."""
    user_email: str
    """Email of the user who uploaded the executable."""
    uploaded_at: datetime
    """Time the executable was uploaded."""
    requirements: VersionExecutableRequirements
    """Requirements for the executable."""


class Version(BaseModel):
    """A version of an application representing a code artifact or a compiled binary."""

    id: str
    """ID of the version."""
    application_id: str
    """ID of the application that this is a version of."""
    name: str
    """Name of the version."""
    description: str
    """Description of the version."""
    executable: VersionExecutable
    """Executable for the version."""
    created_at: datetime
    """Creation time of the version."""
    updated_at: datetime
    """Last update time of the version."""



"""Module for handling input sources and data."""

import copy
import csv
import json
import os
import sys
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Union

from nextmv.options import Options


class InputFormat(str, Enum):
    """Format of an `Input`."""

    JSON = "JSON"
    """JSON format, utf-8 encoded."""
    TEXT = "TEXT"
    """Text format, utf-8 encoded."""
    CSV = "CSV"
    """CSV format, utf-8 encoded."""
    CSV_ARCHIVE = "CSV_ARCHIVE"
    """CSV archive format: multiple CSV files."""


@dataclass
class Input:
    """
    Input for a decision problem.

    Parameters
    ----------
    data : Any
        The actual data.
    input_format : InputFormat, optional
        Format of the input data. Default is `InputFormat.JSON`.
    options : Options, optional
        Options that the input was created with.
    """

    data: Union[
        Union[dict[str, Any], Any],  # JSON
        str,  # TEXT
        list[dict[str, Any]],  # CSV
        dict[str, list[dict[str, Any]]],  # CSV_ARCHIVE
    ]
    """The actual data. The data can be of various types, depending on the
    input format."""

    input_format: Optional[InputFormat] = InputFormat.JSON
    """Format of the input data. Default is `InputFormat.JSON`."""
    options: Optional[Options] = None
    """Options that the `Input` were created with."""

    def __post_init__(self):
        """Check that the data matches the format given to initialize the
        class."""

        if self.input_format == InputFormat.JSON:
            try:
                _ = json.dumps(self.data)
            except (TypeError, OverflowError) as e:
                raise ValueError(
                    f"Input has input_format InputFormat.JSON and "
                    f"data is of type {type(self.data)}, which is not JSON serializable"
                ) from e

        elif self.input_format == InputFormat.TEXT and not isinstance(self.data, str):
            raise ValueError(
                f"unsupported Input.data type: {type(self.data)} with "
                "input_format InputFormat.TEXT, supported type is `str`"
            )

        elif self.input_format == InputFormat.CSV and not isinstance(self.data, list):
            raise ValueError(
                f"unsupported Input.data type: {type(self.data)} with "
                "input_format InputFormat.CSV, supported type is `list`"
            )

        elif self.input_format == InputFormat.CSV_ARCHIVE and not isinstance(self.data, dict):
            raise ValueError(
                f"unsupported Input.data type: {type(self.data)} with "
                "input_format InputFormat.CSV_ARCHIVE, supported type is `dict`"
            )

        # Capture a snapshot of the options that were used to create the class
        # so even if they are changed later, we have a record of the original.
        init_options = self.options
        new_options = copy.deepcopy(init_options)
        self.options = new_options


class InputLoader:
    """Base class for loading inputs."""

    def load(
        self,
        input_format: InputFormat = InputFormat.JSON,
        options: Optional[Options] = None,
        *args,
        **kwargs,
    ) -> Input:
        """
        Read the input data. This method should be implemented by
        subclasses.

        Parameters
        ----------
        input_format : InputFormat, optional
            Format of the input data. Default is `InputFormat.JSON`.
        options : Options, optional
            Options for loading the input data.
        *args
            Additional positional arguments.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        Input
            The input data.

        Raises
        ------
        NotImplementedError
            If the method is not implemented.
        """

        raise NotImplementedError


class LocalInputLoader(InputLoader):
    """
    Class for loading local inputs. This class can load input data from the
    local filesystem, by using stdin, a file, or a directory, where applicable.
    Call the `load` method to read the input data.
    """

    def _read_text(path: str, _) -> str:
        with open(path, encoding="utf-8") as f:
            return f.read().rstrip("\n")

    def _read_csv(path: str, csv_configurations: Optional[dict[str, Any]]) -> list[dict[str, Any]]:
        with open(path, encoding="utf-8") as f:
            return list(csv.DictReader(f, **csv_configurations))

    def _read_json(path: str, _) -> Union[dict[str, Any], Any]:
        with open(path, encoding="utf-8") as f:
            return json.load(f)

    # All of these readers are callback functions.
    STDIN_READERS = {
        InputFormat.JSON: lambda _: json.load(sys.stdin),
        InputFormat.TEXT: lambda _: sys.stdin.read().rstrip("\n"),
        InputFormat.CSV: lambda csv_configurations: list(csv.DictReader(sys.stdin, **csv_configurations)),
    }
    # These callbacks were not implemented with lambda because we needed
    # multiple lines. By using `open`, we needed the `with` to be able to close
    # the file.
    FILE_READERS = {
        InputFormat.JSON: _read_json,
        InputFormat.TEXT: _read_text,
        InputFormat.CSV: _read_csv,
    }

    def load(
        self,
        input_format: Optional[InputFormat] = InputFormat.JSON,
        options: Optional[Options] = None,
        path: Optional[str] = None,
        csv_configurations: Optional[dict[str, Any]] = None,
    ) -> Input:
        """
        Load the input data. The input data can be in various formats. For
        `InputFormat.JSON`, `InputFormat.TEXT`, and `InputFormat.CSV`, the data
        can be streamed from stdin or read from a file. When the `path`
        argument is provided (and valid), the input data is read from the file
        specified by `path`, otherwise, it is streamed from stdin. For
        `InputFormat.CSV_ARCHIVE`, the input data is read from the directory
        specified by `path`. If the `path` is not provided, the default
        location `input` is used. The directory should contain one or more
        files, where each file in the directory is a CSV file.

        The `Input` that is returned contains the `data` attribute. This data
        can be of different types, depending on the provided `input_format`:

        - `InputFormat.JSON`: the data is a `dict[str, Any]`.
        - `InputFormat.TEXT`: the data is a `str`.
        - `InputFormat.CSV`: the data is a `list[dict[str, Any]]`.
        - `InputFormat.CSV_ARCHIVE`: the data is a `dict[str, list[dict[str, Any]]]`.
          Each key is the name of the CSV file, minus the `.csv` extension.

        Parameters
        ----------
        input_format : InputFormat, optional
            Format of the input data. Default is `InputFormat.JSON`.
        options : Options, optional
            Options for loading the input data.
        path : str, optional
            Path to the input data.
        csv_configurations : dict[str, Any], optional
            Configurations for loading CSV files. The default `DictReader` is
            used when loading a CSV file, so you have the option to pass in a
            dictionary with custom kwargs for the `DictReader`.

        Returns
        -------
        Input
            The input data.

        Raises
        ------
        ValueError
            If the path is not a directory when working with CSV_ARCHIVE.
        """

        data: Any = None
        if csv_configurations is None:
            csv_configurations = {}

        if input_format in [InputFormat.JSON, InputFormat.TEXT, InputFormat.CSV]:
            data = self._load_utf8_encoded(path=path, input_format=input_format, csv_configurations=csv_configurations)
        elif input_format == InputFormat.CSV_ARCHIVE:
            data = self._load_archive(path=path, csv_configurations=csv_configurations)

        return Input(data=data, input_format=input_format, options=options)

    def _load_utf8_encoded(
        self,
        csv_configurations: Optional[dict[str, Any]],
        path: Optional[str] = None,
        input_format: Optional[InputFormat] = InputFormat.JSON,
        use_file_reader: bool = False,
    ) -> Union[dict[str, Any], str, list[dict[str, Any]]]:
        """
        Load a utf-8 encoded file. Can come from stdin or a file in the
        filesystem.
        """

        # If we forcibly want to use the file reader, we can do so.
        if use_file_reader:
            return self.FILE_READERS[input_format](path, csv_configurations)

        # Otherwise, we can use the stdin reader if no path is provided.
        if path is None or path == "":
            return self.STDIN_READERS[input_format](csv_configurations)

        # Lastly, we can use the file reader if a path is provided.
        return self.FILE_READERS[input_format](path, csv_configurations)

    def _load_archive(
        self,
        csv_configurations: Optional[dict[str, Any]],
        path: Optional[str] = None,
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Load files from a directory. Will only load CSV files.
        """

        dir_path = "input"
        if path is not None and path != "":
            if not os.path.isdir(path):
                raise ValueError(f"path {path} is not a directory")

            dir_path = path

        if not os.path.isdir(dir_path):
            raise ValueError(f'expected input directoy "{dir_path}" to exist as a default location')

        data = {}
        csv_ext = ".csv"
        for file in os.listdir(dir_path):
            if file.endswith(csv_ext):
                stripped = file.removesuffix(csv_ext)
                data[stripped] = self._load_utf8_encoded(
                    path=os.path.join(dir_path, file),
                    input_format=InputFormat.CSV,
                    use_file_reader=True,
                    csv_configurations=csv_configurations,
                )

        return data


def load_local(
    input_format: Optional[InputFormat] = InputFormat.JSON,
    options: Optional[Options] = None,
    path: Optional[str] = None,
    csv_configurations: Optional[dict[str, Any]] = None,
) -> Input:
    """
    This is a convenience function for instantiating a `LocalInputLoader`
    and calling its `load` method.

    Load the input data. The input data can be in various formats. For
    `InputFormat.JSON`, `InputFormat.TEXT`, and `InputFormat.CSV`, the data can
    be streamed from stdin or read from a file. When the `path` argument is
    provided (and valid), the input data is read from the file specified by
    `path`, otherwise, it is streamed from stdin. For
    `InputFormat.CSV_ARCHIVE`, the input data is read from the directory
    specified by `path`. If the `path` is not provided, the default location
    `input` is used. The directory should contain one or more files, where each
    file in the directory is a CSV file.

    The `Input` that is returned contains the `data` attribute. This data can
    be of different types, depending on the provided `input_format`:

    - `InputFormat.JSON`: the data is a `dict[str, Any]`.
    - `InputFormat.TEXT`: the data is a `str`.
    - `InputFormat.CSV`: the data is a `list[dict[str, Any]]`.
    - `InputFormat.CSV_ARCHIVE`: the data is a `dict[str, list[dict[str, Any]]]`.
        Each key is the name of the CSV file, minus the `.csv` extension.

    Parameters
    ----------
    input_format : InputFormat, optional
        Format of the input data. Default is `InputFormat.JSON`.
    options : Options, optional
        Options for loading the input data.
    path : str, optional
        Path to the input data.
    csv_configurations : dict[str, Any], optional
        Configurations for loading CSV files. The default `DictReader` is used
        when loading a CSV file, so you have the option to pass in a dictionary
        with custom kwargs for the `DictReader`.

    Returns
    -------
    Input
        The input data.

    Raises
    ------
    ValueError
        If the path is not a directory when working with CSV_ARCHIVE.
    """

    loader = LocalInputLoader()
    return loader.load(input_format, options, path, csv_configurations)



"""Module with the logic for handling an app manifest."""

import os
from enum import Enum
from typing import Any, Optional

import yaml
from pydantic import Field

from nextmv.base_model import BaseModel
from nextmv.model import _REQUIREMENTS_FILE, ModelConfiguration

FILE_NAME = "app.yaml"
"""Name of the app manifest file."""


class ManifestType(str, Enum):
    """Type of application in the manifest, based on the programming
    language."""

    PYTHON = "python"
    """Python format"""
    GO = "go"
    """Go format"""
    JAVA = "java"
    """Java format"""


class ManifestRuntime(str, Enum):
    """Runtime (environment) where the app will be run on Nextmv Cloud."""

    DEFAULT = "ghcr.io/nextmv-io/runtime/default:latest"
    """This runtime is used to run compiled applications such as Go binaries."""
    PYTHON = "ghcr.io/nextmv-io/runtime/python:3.11"
    """
    This runtime is used as the basis for all other Python runtimes and Python
    applications.
    """
    JAVA = "ghcr.io/nextmv-io/runtime/java:latest"
    """This runtime is used to run Java applications."""
    PYOMO = "ghcr.io/nextmv-io/runtime/pyomo:latest"
    """This runtime provisions Python packages to run Pyomo applications."""
    HEXALY = "ghcr.io/nextmv-io/runtime/hexaly:latest"
    """
    Based on the python runtime, it provisions (pre-installs) the Hexaly solver
    to run Python applications.
    """


class ManifestBuild(BaseModel):
    """Build-specific attributes."""

    command: Optional[str] = None
    """
    The command to run to build the app. This command will be executed without
    a shell, i.e., directly. The command must exit with a status of 0 to
    continue the push process of the app to Nextmv Cloud. This command is
    executed prior to the pre-push command.
    """
    environment: Optional[dict[str, Any]] = None
    """
    Environment variables to set when running the build command given as
    key-value pairs.
    """

    def environment_to_dict(self) -> dict[str, str]:
        """
        Convert the environment variables to a dictionary.

        Returns
        -------
        dict[str, str]
            The environment variables as a dictionary.

        """

        if self.environment is None:
            return {}

        return {key: str(value) for key, value in self.environment.items()}


class ManifestPythonModel(BaseModel):
    """Model-specific instructions for a Python app."""

    name: str
    """The name of the decision model."""
    options: Optional[list[dict[str, Any]]] = None
    """
    Options for the decision model. This is a data representation of the
    `nextmv.Options` class. It consists of a list of dicts. Each dict
    represents the `nextmv.Parameter` class. It is used to be able to
    reconstruct an Options object from data when loading a decision model.
    """


class ManifestPython(BaseModel):
    """Python-specific instructions."""

    pip_requirements: Optional[str] = Field(serialization_alias="pip-requirements", default=None)
    """
    Path to a requirements.txt file containing (additional) Python
    dependencies that will be bundled with the app.
    """
    model: Optional[ManifestPythonModel] = None
    """
    Information about an encoded decision model as handlded via mlflow. This
    information is used to load the decision model from the app bundle.
    """


class Manifest(BaseModel):
    """
    An application that runs on the Nextmv Platform must contain a file named
    `app.yaml` which is known as the app manifest. This file is used to specify
    the execution environment for the app.

    This class represents the app manifest and allows you to load it from a
    file or create it programmatically.
    """

    files: list[str]
    """Mandatory. The files to include (or exclude) in the app."""

    runtime: ManifestRuntime = ManifestRuntime.PYTHON
    """
    Mandatory. The runtime to use for the app, it provides the environment in
    which the app runs.
    """
    type: ManifestType = ManifestType.PYTHON
    """Mandatory. Type of application, based on the programming language."""
    build: Optional[ManifestBuild] = None
    """
    Optional. Build-specific attributes. The build.command to run to build the
    app. This command will be executed without a shell, i.e., directly. The
    command must exit with a status of 0 to continue the push process of the
    app to Nextmv Cloud. This command is executed prior to the pre-push
    command. The build.environment is used to set environment variables when
    running the build command given as key-value pairs.
    """
    pre_push: Optional[str] = Field(serialization_alias="pre-push", default=None)
    """
    Optional. A command to run before the app is pushed to the Nextmv Cloud.
    This command can be used to compile a binary, run tests or similar tasks.
    One difference with what is specified under build, is that the command will
    be executed via the shell (i.e., bash -c on Linux & macOS or cmd /c on
    Windows). The command must exit with a status of 0 to continue the push
    process. This command is executed just before the app gets bundled and
    pushed (after the build command).
    """
    python: Optional[ManifestPython] = None
    """
    Optional. Only for Python apps. Contains further Python-specific
    attributes.
    """

    @classmethod
    def from_yaml(cls, dirpath: str) -> "Manifest":
        """
        Load a manifest from a YAML file.

        Parameters
        ----------
        dirpath : str
            Path to the directory containing the app.yaml file.

        Returns
        -------
        Manifest
            The loaded manifest.

        """

        with open(os.path.join(dirpath, FILE_NAME)) as file:
            raw_manifest = yaml.safe_load(file)

        return cls.from_dict(raw_manifest)

    def to_yaml(self, dirpath: str) -> None:
        """
        Write the manifest to a YAML file.

        Parameters
        ----------
        dirpath : str
            Path to the directory where the app.yaml file will be written.

        """

        with open(os.path.join(dirpath, FILE_NAME), "w") as file:
            yaml.dump(self.to_dict(), file)

    @classmethod
    def from_model_configuration(cls, model_configuration: ModelConfiguration) -> "Manifest":
        """
        Create a Python manifest from a Python model configuration.

        Parameters
        ----------
        model_configuration : ModelConfiguration
            The model configuration.

        Returns
        -------
        Manifest
            The Python manifest.
        """

        manifest_python_dict = {
            "pip-requirements": _REQUIREMENTS_FILE,
            "model": {
                "name": model_configuration.name,
            },
        }

        if model_configuration.options is not None:
            manifest_python_dict["model"]["options"] = model_configuration.options.parameters_dict()

        manifest_python = ManifestPython.from_dict(manifest_python_dict)

        return cls(
            files=["main.py", f"{model_configuration.name}/**"],
            runtime=ManifestRuntime.PYTHON,
            type=ManifestType.PYTHON,
            python=manifest_python,
        )



"""JSON class for data wrangling JSON objects."""

from importlib import import_module
from typing import Any, Optional

from pydantic import BaseModel


class BaseModel(BaseModel):
    """Base class for data wrangling tasks with JSON."""

    @classmethod
    def from_dict(cls, data: Optional[dict[str, Any]] = None):
        """Instantiates the class from a dict."""

        if data is None:
            return None

        return cls(**data)

    def to_dict(self) -> dict[str, Any]:
        """Converts the class to a dict."""

        return self.model_dump(mode="json", exclude_none=True, by_alias=True)


def from_dict(data: dict[str, Any]) -> Any:
    """
    Load a data model instance from a dict with associated class info.

    Parameters
    ----------
    data : dict[str, Any]
        The data to load.

    Returns
    -------
    Any
        The loaded data model instance.
    """

    module = import_module(data["class"]["module"])
    cls = getattr(module, data["class"]["name"])

    return cls.from_dict(data["attributes"])
